# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Blackbox tests for version_info."""

import os

from breezy.tests import TestCaseWithTransport
from breezy.version_info_formats import VersionInfoBuilder


class TestVersionInfo(TestCaseWithTransport):
    def test_invalid_format(self):
        self.run_bzr("version-info --format quijibo", retcode=3)

    def create_tree(self):
        wt = self.make_branch_and_tree("branch")

        self.build_tree(["branch/a"])
        wt.add("a")
        wt.commit("adding a", rev_id=b"r1")

        self.build_tree(["branch/b"])
        wt.add("b")
        wt.commit("adding b", rev_id=b"r2")
        return wt

    def test_basic(self):
        wt = self.create_tree()

        txt = self.run_bzr("version-info branch")[0]
        self.assertContainsRe(txt, "date:")
        self.assertContainsRe(txt, "build-date:")
        self.assertContainsRe(txt, "revno: 2")
        self.assertContainsRe(
            txt, "revision-id: " + wt.branch.last_revision().decode("utf-8")
        )

    def test_all(self):
        """'--all' includes clean, revision history, and file revisions."""
        wt = self.create_tree()
        txt = self.run_bzr("version-info branch --all")[0]
        self.assertContainsRe(txt, "date:")
        self.assertContainsRe(txt, "revno: 2")
        self.assertContainsRe(
            txt, "revision-id: " + wt.branch.last_revision().decode("utf-8")
        )
        self.assertContainsRe(txt, "clean: True")
        self.assertContainsRe(txt, "revisions:")
        for rev_id in wt.branch.repository.all_revision_ids():
            self.assertContainsRe(txt, "id: " + rev_id.decode("utf-8"))
        self.assertContainsRe(txt, "message: adding a")
        self.assertContainsRe(txt, "message: adding b")
        self.assertContainsRe(txt, "file-revisions:")
        self.assertContainsRe(txt, "path: a")
        self.assertContainsRe(txt, "path: b")

    def test_clean(self):
        """Test that --check-clean includes the right info."""
        self.create_tree()

        txt = self.run_bzr("version-info branch --check-clean")[0]
        self.assertContainsRe(txt, "clean: True")

        self.build_tree_contents([("branch/c", b"now unclean\n")])
        txt = self.run_bzr("version-info branch --check-clean")[0]
        self.assertContainsRe(txt, "clean: False")

        txt = self.run_bzr(
            "version-info branch --check-clean --include-file-revisions"
        )[0]
        self.assertContainsRe(txt, "revision: unversioned")

        os.remove("branch/c")

    def test_no_working_tree(self):
        tree = self.create_tree()
        branch = self.make_branch("just_branch")
        branch.pull(tree.branch)

        txt = self.run_bzr("version-info just_branch")[0]
        self.assertStartsWith(txt, "revision-id: r2\n")

    def assertEqualNoBuildDate(self, text1, text2):
        """Compare 2 texts, but ignore the build-date field.

        build-date is the current timestamp, accurate to seconds. But the
        clock is always ticking, and it may have ticked between the time
        that text1 and text2 were generated.
        """
        lines1 = text1.splitlines(True)
        lines2 = text2.splitlines(True)
        for line1, line2 in zip(lines1, lines2):
            if line1.startswith("build-date: "):
                self.assertStartsWith(line2, "build-date: ")
            else:
                self.assertEqual(line1, line2)
        self.assertEqual(len(lines1), len(lines2))

    def test_no_branch(self):
        """Test that brz defaults to the local working directory."""
        self.create_tree()

        txt1 = self.run_bzr("version-info branch")[0]

        os.chdir("branch")
        txt2 = self.run_bzr("version-info")[0]
        self.assertEqualNoBuildDate(txt1, txt2)

    def test_rio(self):
        """Test that we can pass --format=rio."""
        self.create_tree()

        txt = self.run_bzr("version-info branch")[0]
        txt1 = self.run_bzr("version-info --format rio branch")[0]
        txt2 = self.run_bzr("version-info --format=rio branch")[0]
        self.assertEqualNoBuildDate(txt, txt1)
        self.assertEqualNoBuildDate(txt, txt2)

    def test_python(self):
        """Test that we can do --format=python."""
        self.create_tree()

        txt = self.run_bzr("version-info --format python branch")[0]

        self.assertContainsRe(txt, "version_info = {")

    def test_custom_without_template(self):
        self.make_branch_and_tree("branch")
        out, err = self.run_bzr("version-info --custom", retcode=3)
        self.assertContainsRe(err, r"ERROR: No template specified\.")

    def test_custom_implies_all(self):
        self.create_tree()
        out, err = self.run_bzr(
            'version-info --custom --template="{revno} {branch_nick} {clean}\n" branch'
        )
        self.assertEqual("2 branch 1\n", out)
        self.assertEqual("", err)
        self.build_tree_contents([("branch/c", b"now unclean\n")])
        out, err = self.run_bzr(
            'version-info --custom --template="{revno} {branch_nick} {clean}\n" branch'
        )
        self.assertEqual("2 branch 0\n", out)
        self.assertEqual("", err)

    def test_custom_no_clean_in_template(self):
        def should_not_be_called(self):
            raise AssertionError(
                "Method on {!r} should not have been used".format(self)
            )

        self.overrideAttr(
            VersionInfoBuilder, "_extract_file_revisions", should_not_be_called
        )
        self.create_tree()
        out, err = self.run_bzr("version-info --custom --template=r{revno} branch")
        self.assertEqual("r2", out)
        self.assertEqual("", err)

    def test_non_ascii(self):
        """Test that we can output non-ascii data."""
        commit_message = "Non-ascii message with character not in latin-1: \u1234"

        tree = self.make_branch_and_tree(".")
        self.build_tree(["a_file"])
        tree.add("a_file")
        tree.commit(commit_message)
        out, err = self.run_bzr_raw(
            ["version-info", "--include-history"], encoding="latin-1"
        )

    def test_revision(self):
        tree = self.create_tree()
        branch = self.make_branch("just_branch")
        branch.pull(tree.branch)

        txt = self.run_bzr("version-info -r1 just_branch")[0]
        self.assertStartsWith(txt, "revision-id: r1\n")



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the osutils wrapper."""

import codecs
import locale
import sys

from .. import osutils
from . import TestCase
from .ui_testing import StringIOWithEncoding


class FakeCodec:
    """Special class that helps testing over several non-existed encodings.

    Clients can add new encoding names, but because of how codecs is
    implemented they cannot be removed. Be careful with naming to avoid
    collisions between tests.
    """

    _registered: bool = False
    _enabled_encodings: set[str] = set()

    def add(self, encoding_name):
        """Adding encoding name to fake.

        :type   encoding_name:  lowercase plain string
        """
        if not self._registered:
            codecs.register(self)
            self._registered = True
        if encoding_name is not None:
            self._enabled_encodings.add(encoding_name)

    def __call__(self, encoding_name):
        """Called indirectly by codecs module during lookup."""
        if encoding_name in self._enabled_encodings:
            return codecs.lookup("latin-1")


fake_codec = FakeCodec()


class TestFakeCodec(TestCase):
    def test_fake_codec(self):
        self.assertRaises(LookupError, codecs.lookup, "fake")

        fake_codec.add("fake")
        codecs.lookup("fake")


class TestTerminalEncoding(TestCase):
    """Test the auto-detection of proper terminal encoding."""

    def setUp(self):
        super().setUp()
        self.overrideAttr(sys, "stdin")
        self.overrideAttr(sys, "stdout")
        self.overrideAttr(sys, "stderr")
        self.overrideAttr(osutils, "_cached_user_encoding")

    def make_wrapped_streams(
        self,
        stdout_encoding,
        stderr_encoding,
        stdin_encoding,
        user_encoding="user_encoding",
        enable_fake_encodings=True,
    ):
        sys.stdout = StringIOWithEncoding()
        sys.stdout.encoding = stdout_encoding
        sys.stderr = StringIOWithEncoding()
        sys.stderr.encoding = stderr_encoding
        sys.stdin = StringIOWithEncoding()
        sys.stdin.encoding = stdin_encoding
        osutils._cached_user_encoding = user_encoding
        if enable_fake_encodings:
            fake_codec.add(stdout_encoding)
            fake_codec.add(stderr_encoding)
            fake_codec.add(stdin_encoding)

    def test_get_terminal_encoding(self):
        self.make_wrapped_streams(
            "stdout_encoding", "stderr_encoding", "stdin_encoding"
        )

        # first preference is stdout encoding
        self.assertEqual("stdout_encoding", osutils.get_terminal_encoding())

        sys.stdout.encoding = None
        # if sys.stdout is None, fall back to sys.stdin
        self.assertEqual("stdin_encoding", osutils.get_terminal_encoding())

        sys.stdin.encoding = None
        # and in the worst case, use osutils.get_user_encoding()
        self.assertEqual("user_encoding", osutils.get_terminal_encoding())

    def test_get_terminal_encoding_silent(self):
        self.make_wrapped_streams(
            "stdout_encoding", "stderr_encoding", "stdin_encoding"
        )
        # Calling get_terminal_encoding should not mutter when silent=True is
        # passed.
        log = self.get_log()
        osutils.get_terminal_encoding()
        self.assertEqual(log, self.get_log())

    def test_get_terminal_encoding_trace(self):
        self.make_wrapped_streams(
            "stdout_encoding", "stderr_encoding", "stdin_encoding"
        )
        # Calling get_terminal_encoding should not mutter when silent=True is
        # passed.
        log = self.get_log()
        osutils.get_terminal_encoding(trace=True)
        self.assertNotEqual(log, self.get_log())

    def test_terminal_cp0(self):
        # test cp0 encoding (Windows returns cp0 when there is no encoding)
        self.make_wrapped_streams(
            "cp0", "cp0", "cp0", user_encoding="latin-1", enable_fake_encodings=False
        )

        # cp0 is invalid encoding. We should fall back to user_encoding
        self.assertEqual("latin-1", osutils.get_terminal_encoding())

        # check stderr
        self.assertEqual("", sys.stderr.getvalue())

    def test_terminal_cp_unknown(self):
        # test against really unknown encoding
        # catch warning at stderr
        self.make_wrapped_streams(
            "cp-unknown",
            "cp-unknown",
            "cp-unknown",
            user_encoding="latin-1",
            enable_fake_encodings=False,
        )

        self.assertEqual("latin-1", osutils.get_terminal_encoding())

        # check stderr
        self.assertEqual(
            "brz: warning: unknown terminal encoding cp-unknown.\n"
            "  Using encoding latin-1 instead.\n",
            sys.stderr.getvalue(),
        )


class TestUserEncoding(TestCase):
    """Test detection of default user encoding."""

    def setUp(self):
        super().setUp()
        self.overrideAttr(osutils, "_cached_user_encoding", None)
        self.overrideAttr(locale, "getpreferredencoding", self.get_encoding)
        self.overrideAttr(locale, "CODESET", None)
        self.overrideAttr(sys, "stderr", StringIOWithEncoding())

    def get_encoding(self, do_setlocale=True):
        return self._encoding

    def test_get_user_encoding(self):
        self._encoding = "user_encoding"
        fake_codec.add("user_encoding")
        self.assertEqual(
            "iso8859-1",  # fake_codec maps to latin-1
            osutils.get_user_encoding(),
        )
        self.assertEqual("", sys.stderr.getvalue())

    def test_user_cp0(self):
        self._encoding = "cp0"
        self.assertEqual("ascii", osutils.get_user_encoding())
        self.assertEqual("", sys.stderr.getvalue())

    def test_user_cp_unknown(self):
        self._encoding = "cp-unknown"
        self.assertEqual("ascii", osutils.get_user_encoding())
        self.assertEqual(
            "brz: warning: unknown encoding cp-unknown."
            " Continuing with ascii encoding.\n",
            sys.stderr.getvalue(),
        )

    def test_user_empty(self):
        """Running bzr from a vim script gives '' for a preferred locale."""
        self._encoding = ""
        self.assertEqual("ascii", osutils.get_user_encoding())
        self.assertEqual("", sys.stderr.getvalue())



# Copyright (C) 2008, 2009, 2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for 'brz modified', which shows modified files."""

from breezy.tests import TestCaseWithTransport


class TestModified(TestCaseWithTransport):
    def test_modified(self):
        """Test that 'modified' command reports modified files."""
        self._test_modified("a", "a")

    def test_modified_with_spaces(self):
        """Test that 'modified' command reports modified files with spaces in their names quoted."""
        self._test_modified("a filename with spaces", '"a filename with spaces"')

    def _test_modified(self, name, output):
        def check_modified(expected, null=False):
            command = "modified"
            if null:
                command += " --null"
            out, err = self.run_bzr(command)
            self.assertEqual(out, expected)
            self.assertEqual(err, "")

        # in empty directory, nothing modified
        tree = self.make_branch_and_tree(".")
        check_modified("")

        # with unknown file, still nothing modified
        self.build_tree_contents([(name, b"contents of %s\n" % (name.encode("utf-8")))])
        check_modified("")

        # after add, not modified
        tree.add(name)
        check_modified("")

        # after commit, not modified
        tree.commit(message="add {}".format(output))
        check_modified("")

        # modify the file
        self.build_tree_contents([(name, b"changed\n")])
        check_modified(output + "\n")

        # check null seps - use the unquoted raw name here
        check_modified(name + "\0", null=True)

        # now commit the file and it's no longer modified
        tree.commit(message="modified {}".format(name))
        check_modified("")

    def test_modified_directory(self):
        """Test --directory option."""
        tree = self.make_branch_and_tree("a")
        self.build_tree(["a/README"])
        tree.add("README")
        tree.commit("r1")
        self.build_tree_contents([("a/README", b"changed\n")])
        out, err = self.run_bzr(["modified", "--directory=a"])
        self.assertEqual("README\n", out)



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from io import StringIO

from .. import config
from .. import status as _mod_status
from ..revisionspec import RevisionSpec
from ..status import show_pending_merges, show_tree_status
from . import TestCaseWithTransport


class TestStatus(TestCaseWithTransport):
    def test_pending_none(self):
        # Test whether show_pending_merges works in a tree with no commits
        tree = self.make_branch_and_tree("a")
        tree.commit("empty commit")
        tree2 = self.make_branch_and_tree("b")
        # set a left most parent that is not a present commit
        tree2.add_parent_tree_id(b"some-ghost", allow_leftmost_as_ghost=True)
        # do a merge
        tree2.merge_from_branch(tree.branch)
        output = StringIO()
        with tree2.lock_read():
            show_pending_merges(tree2, output)
        self.assertContainsRe(output.getvalue(), "empty commit")

    def make_multiple_pending_tree(self):
        config.GlobalStack().set("email", "Joe Foo <joe@foo.com>")
        tree = self.make_branch_and_tree("a")
        tree.commit("commit 1", timestamp=1196796819, timezone=0)
        tree2 = tree.controldir.clone("b").open_workingtree()
        tree.commit("commit 2", timestamp=1196796819, timezone=0)
        tree2.commit("commit 2b", timestamp=1196796819, timezone=0)
        tree3 = tree2.controldir.clone("c").open_workingtree()
        tree2.commit("commit 3b", timestamp=1196796819, timezone=0)
        tree3.commit("commit 3c", timestamp=1196796819, timezone=0)
        tree.merge_from_branch(tree2.branch)
        tree.merge_from_branch(tree3.branch, force=True)
        return tree

    def test_multiple_pending(self):
        tree = self.make_multiple_pending_tree()
        output = StringIO()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        show_pending_merges(tree, output)
        # 2b doesn't appear because it's an ancestor of 3b
        self.assertEqualDiff(
            "pending merge tips: (use -v to see all merge revisions)\n"
            "  Joe Foo 2007-12-04 commit 3b\n"
            "  Joe Foo 2007-12-04 commit 3c\n",
            output.getvalue(),
        )

    def test_multiple_pending_verbose(self):
        tree = self.make_multiple_pending_tree()
        output = StringIO()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        show_pending_merges(tree, output, verbose=True)
        # Even though 2b is in the ancestry of 3c, it should only be displayed
        # under the first merge parent.
        self.assertEqualDiff(
            "pending merges:\n"
            "  Joe Foo 2007-12-04 commit 3b\n"
            "    Joe Foo 2007-12-04 commit 2b\n"
            "  Joe Foo 2007-12-04 commit 3c\n",
            output.getvalue(),
        )

    def test_with_pending_ghost(self):
        """Test when a pending merge is itself a ghost."""
        tree = self.make_branch_and_tree("a")
        tree.commit("first")
        tree.add_parent_tree_id(b"a-ghost-revision")
        tree.lock_read()
        self.addCleanup(tree.unlock)
        output = StringIO()
        show_pending_merges(tree, output)
        self.assertEqualDiff(
            "pending merge tips: (use -v to see all merge revisions)\n"
            "  (ghost) a-ghost-revision\n",
            output.getvalue(),
        )

    def test_pending_with_ghosts(self):
        """Test when a pending merge's ancestry includes ghosts."""
        config.GlobalStack().set("email", "Joe Foo <joe@foo.com>")
        tree = self.make_branch_and_tree("a")
        tree.commit("empty commit")
        tree2 = tree.controldir.clone("b").open_workingtree()
        tree2.commit("a non-ghost", timestamp=1196796819, timezone=0)
        tree2.add_parent_tree_id(b"a-ghost-revision")
        tree2.commit("commit with ghost", timestamp=1196796819, timezone=0)
        tree2.commit("another non-ghost", timestamp=1196796819, timezone=0)
        tree.merge_from_branch(tree2.branch)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        output = StringIO()
        show_pending_merges(tree, output, verbose=True)
        self.assertEqualDiff(
            "pending merges:\n"
            "  Joe Foo 2007-12-04 another non-ghost\n"
            "    Joe Foo 2007-12-04 [merge] commit with ghost\n"
            "    (ghost) a-ghost-revision\n"
            "    Joe Foo 2007-12-04 a non-ghost\n",
            output.getvalue(),
        )

    def tests_revision_to_revision(self):
        """Doing a status between two revision trees should work."""
        tree = self.make_branch_and_tree(".")
        r1_id = tree.commit("one", allow_pointless=True)
        r2_id = tree.commit("two", allow_pointless=True)
        output = StringIO()
        show_tree_status(
            tree,
            to_file=output,
            revision=[
                RevisionSpec.from_string("revid:{}".format(r1_id.decode("utf-8"))),
                RevisionSpec.from_string("revid:{}".format(r2_id.decode("utf-8"))),
            ],
        )
        # return does not matter as long as it did not raise.


class TestHooks(TestCaseWithTransport):
    def test_constructor(self):
        """Check that creating a StatusHooks instance has the right defaults."""
        hooks = _mod_status.StatusHooks()
        self.assertTrue("post_status" in hooks, "post_status not in {}".format(hooks))
        self.assertTrue("pre_status" in hooks, "pre_status not in {}".format(hooks))

    def test_installed_hooks_are_StatusHooks(self):
        """The installed hooks object should be a StatusHooks."""
        # the installed hooks are saved in self._preserved_hooks.
        self.assertIsInstance(
            self._preserved_hooks[_mod_status][1], _mod_status.StatusHooks
        )

    def test_post_status_hook(self):
        """Ensure that post_status hook is invoked with the right args."""
        calls = []
        _mod_status.hooks.install_named_hook("post_status", calls.append, None)
        self.assertLength(0, calls)
        tree = self.make_branch_and_tree(".")
        r1_id = tree.commit("one", allow_pointless=True)
        r2_id = tree.commit("two", allow_pointless=True)
        output = StringIO()
        show_tree_status(
            tree,
            to_file=output,
            revision=[
                RevisionSpec.from_string("revid:{}".format(r1_id.decode("utf-8"))),
                RevisionSpec.from_string("revid:{}".format(r2_id.decode("utf-8"))),
            ],
        )
        self.assertLength(1, calls)
        params = calls[0]
        self.assertIsInstance(params, _mod_status.StatusHookParams)
        attrs = [
            "old_tree",
            "new_tree",
            "to_file",
            "versioned",
            "show_ids",
            "short",
            "verbose",
            "specific_files",
        ]
        for a in attrs:
            self.assertTrue(
                hasattr(params, a),
                'Attribute "{}" not found in StatusHookParam'.format(a),
            )

    def test_pre_status_hook(self):
        """Ensure that pre_status hook is invoked with the right args."""
        calls = []
        _mod_status.hooks.install_named_hook("pre_status", calls.append, None)
        self.assertLength(0, calls)
        tree = self.make_branch_and_tree(".")
        r1_id = tree.commit("one", allow_pointless=True)
        r2_id = tree.commit("two", allow_pointless=True)
        output = StringIO()
        show_tree_status(
            tree,
            to_file=output,
            revision=[
                RevisionSpec.from_string("revid:{}".format(r1_id.decode("utf-8"))),
                RevisionSpec.from_string("revid:{}".format(r2_id.decode("utf-8"))),
            ],
        )
        self.assertLength(1, calls)
        params = calls[0]
        self.assertIsInstance(params, _mod_status.StatusHookParams)
        attrs = [
            "old_tree",
            "new_tree",
            "to_file",
            "versioned",
            "show_ids",
            "short",
            "verbose",
            "specific_files",
        ]
        for a in attrs:
            self.assertTrue(
                hasattr(params, a),
                'Attribute "{}" not found in StatusHookParam'.format(a),
            )



#!/usr/bin/python3

# Copyright 2005 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""%(prog)s - generate information from built-in brz help.

%(prog)s creates a file with information on brz in one of
several different output formats:

    man              man page
    bash_completion  bash completion script
    ...

Examples:
    python generated-docs.py man
    python generated-docs.py bash_completion

Run "%(prog)s --help" for the option reference.
"""

import os
import sys
from optparse import OptionParser

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

from contextlib import ExitStack

import breezy
from breezy import commands, doc_generate


def main(argv):
    parser = OptionParser(
        usage="""%prog [options] OUTPUT_FORMAT

Available OUTPUT_FORMAT:

    man              man page
    rstx             man page in ReStructuredText format
    bash_completion  bash completion script"""
    )

    parser.add_option(
        "-s",
        "--show-filename",
        action="store_true",
        dest="show_filename",
        default=False,
        help="print default filename on stdout",
    )

    parser.add_option(
        "-o", "--output", dest="filename", metavar="FILE", help="write output to FILE"
    )

    parser.add_option(
        "-b",
        "--brz-name",
        dest="brz_name",
        default="brz",
        metavar="EXEC_NAME",
        help="name of brz executable",
    )

    parser.add_option(
        "-e",
        "--examples",
        action="callback",
        callback=print_extended_help,
        help="Examples of ways to call generate_doc",
    )

    (options, args) = parser.parse_args(argv)

    if len(args) != 2:
        parser.print_help()
        sys.exit(1)

    with breezy.initialize(), ExitStack() as es:
        # Import breezy.bzr for format registration, see <http://pad.lv/956860>
        from breezy import bzr as _  # noqa: F401

        commands.install_bzr_command_hooks()
        infogen_type = args[1]
        infogen_mod = doc_generate.get_module(infogen_type)
        if options.filename:
            outfilename = options.filename
        else:
            outfilename = infogen_mod.get_filename(options)
        if outfilename == "-":
            outfile = sys.stdout
        else:
            outfile = es.enter_context(open(outfilename, "w"))
        if options.show_filename and (outfilename != "-"):
            sys.stdout.write(outfilename)
            sys.stdout.write("\n")
        infogen_mod.infogen(options, outfile)


def print_extended_help(option, opt, value, parser):
    """Program help examples.

    Prints out the examples stored in the docstring.

    """
    sys.stdout.write(__doc__ % {"prog": sys.argv[0]})
    sys.stdout.write("\n")
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv)



# Copyright (C) 2006, 2007, 2009, 2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
"""Black-box tests for brz aliases."""

from breezy import config
from breezy.tests import TestCaseWithTransport


class TestAliases(TestCaseWithTransport):
    def test_aliases(self):
        def bzr(args, **kwargs):
            return self.run_bzr(args, **kwargs)[0]

        def bzr_catch_error(args, **kwargs):
            return self.run_bzr(args, **kwargs)[1]

        config.GlobalConfig.from_string(
            b"""[ALIASES]
c=cat
c1=cat -r 1
c2=cat -r 1 -r2
""",
            save=True,
        )

        str1 = "foo\n"
        str2 = "bar\n"

        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("a", str1)])
        tree.add("a")
        tree.commit(message="1")

        self.assertEqual(bzr("c a"), str1)

        self.build_tree_contents([("a", str2)])
        tree.commit(message="2")

        self.assertEqual(bzr("c a"), str2)
        self.assertEqual(bzr("c1 a"), str1)
        self.assertEqual(bzr("c1 --revision 2 a"), str2)

        # If --no-aliases isn't working, we will not get retcode=3
        bzr("--no-aliases c a", retcode=3)

        # If --no-aliases breaks all of bzr, we also get retcode=3
        # So we need to catch the output as well
        self.assertEqual(
            bzr_catch_error("--no-aliases c a", retcode=None),
            'brz: ERROR: unknown command "c". Perhaps you meant "ci"\n',
        )

        bzr("c -r1 -r2", retcode=3)
        bzr("c1 -r1 -r2", retcode=3)
        bzr("c2", retcode=3)
        bzr("c2 -r1", retcode=3)



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Test 'brz init'."""

import os
import re

from breezy import branch as _mod_branch
from breezy import config as _mod_config
from breezy import osutils, urlutils
from breezy.bzr.bzrdir import BzrDirMetaFormat1
from breezy.tests import TestCaseWithTransport, TestSkipped
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer
from breezy.workingtree import WorkingTree


class TestInit(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self._default_label = "2a"

    def test_init_with_format(self):
        # Verify brz init --format constructs something plausible
        t = self.get_transport()
        self.run_bzr("init --format default")
        self.assertIsDirectory(".bzr", t)
        self.assertIsDirectory(".bzr/checkout", t)
        self.assertIsDirectory(".bzr/checkout/lock", t)

    def test_init_format_2a(self):
        """Smoke test for constructing a format 2a repository."""
        out, err = self.run_bzr("init --format=2a")
        self.assertEqual("""Created a standalone tree (format: 2a)\n""", out)
        self.assertEqual("", err)

    def test_init_format_bzr(self):
        """Smoke test for constructing a format with the 'bzr' alias."""
        out, err = self.run_bzr("init --format=bzr")
        self.assertEqual(
            "Created a standalone tree (format: {})\n".format(self._default_label), out
        )
        self.assertEqual("", err)

    def test_init_colocated(self):
        """Smoke test for constructing a colocated branch."""
        out, err = self.run_bzr("init --format=development-colo file:,branch=abranch")
        self.assertEqual(
            """Created a standalone tree (format: development-colo)\n""", out
        )
        self.assertEqual("", err)
        out, err = self.run_bzr("branches")
        self.assertEqual("  abranch\n", out)
        self.assertEqual("", err)

    def test_init_at_repository_root(self):
        # brz init at the root of a repository should create a branch
        # and working tree even when creation of working trees is disabled.
        t = self.get_transport()
        t.mkdir("repo")
        format = BzrDirMetaFormat1()
        newdir = format.initialize(t.abspath("repo"))
        repo = newdir.create_repository(shared=True)
        repo.set_make_working_trees(False)
        out, err = self.run_bzr("init repo")
        self.assertEqual(
            """Created a repository tree (format: {})
Using shared repository: {}
""".format(
                self._default_label,
                urlutils.local_path_from_url(
                    repo.controldir.root_transport.external_url()
                ),
            ),
            out,
        )
        cwd = osutils.getcwd()
        self.assertEndsWith(out, cwd + "/repo/\n")
        self.assertEqual("", err)
        newdir.open_branch()
        newdir.open_workingtree()

    def test_init_branch(self):
        out, err = self.run_bzr("init")
        self.assertEqual(
            "Created a standalone tree (format: {})\n".format(self._default_label), out
        )
        self.assertEqual("", err)

        # Can it handle subdirectories of branches too ?
        out, err = self.run_bzr("init subdir1")
        self.assertEqual(
            "Created a standalone tree (format: {})\n".format(self._default_label), out
        )
        self.assertEqual("", err)
        WorkingTree.open("subdir1")

        self.run_bzr_error(
            ["Parent directory of subdir2/nothere does not exist"],
            "init subdir2/nothere",
        )
        out, err = self.run_bzr("init subdir2/nothere", retcode=3)
        self.assertEqual("", out)

        os.mkdir("subdir2")
        out, err = self.run_bzr("init subdir2")
        self.assertEqual(
            "Created a standalone tree (format: {})\n".format(self._default_label), out
        )
        self.assertEqual("", err)
        # init an existing branch.
        out, err = self.run_bzr("init subdir2", retcode=3)
        self.assertEqual("", out)
        self.assertTrue(err.startswith("brz: ERROR: Already a branch:"))

    def test_init_branch_quiet(self):
        out, err = self.run_bzr("init -q")
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_init_existing_branch(self):
        self.run_bzr("init")
        out, err = self.run_bzr("init", retcode=3)
        self.assertContainsRe(err, "Already a branch")
        # don't suggest making a checkout, there's already a working tree
        self.assertFalse(re.search(r"checkout", err))

    def test_init_existing_without_workingtree(self):
        # make a repository
        repo = self.make_repository(".", shared=True)
        repo.set_make_working_trees(False)
        # make a branch; by default without a working tree
        self.run_bzr("init subdir")
        # fail
        out, err = self.run_bzr("init subdir", retcode=3)
        # suggests using checkout
        self.assertContainsRe(err, "ontains a branch.*but no working tree.*checkout")

    def test_no_defaults(self):
        """Init creates no default ignore rules."""
        self.run_bzr("init")
        self.assertFalse(os.path.exists(".bzrignore"))

    def test_init_unicode(self):
        # Make sure getcwd can handle unicode filenames
        try:
            os.mkdir("mu-\xb5")
        except UnicodeError:
            raise TestSkipped("Unable to create Unicode filename")
        # try to init unicode dir
        self.run_bzr(["init", "-q", "mu-\xb5"])

    def create_simple_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a"])
        tree.add(["a"], ids=[b"a-id"])
        tree.commit("one", rev_id=b"r1")
        return tree

    def test_init_create_prefix(self):
        """'brz init --create-prefix; will create leading directories."""
        self.create_simple_tree()

        self.run_bzr_error(
            ["Parent directory of ../new/tree does not exist"],
            "init ../new/tree",
            working_dir="tree",
        )
        self.run_bzr("init ../new/tree --create-prefix", working_dir="tree")
        self.assertPathExists("new/tree/.bzr")

    def test_init_default_format_option(self):
        """Brz init should read default format from option default_format."""
        g_store = _mod_config.GlobalStore()
        g_store._load_from_string(b"""
[DEFAULT]
default_format = 1.9
""")
        g_store.save()
        out, err = self.run_brz_subprocess("init")
        self.assertContainsRe(out, b"1.9")

    def test_init_no_tree(self):
        """'brz init --no-tree' creates a branch with no working tree."""
        out, err = self.run_bzr("init --no-tree")
        self.assertStartsWith(out, "Created a standalone branch")


class TestSFTPInit(TestCaseWithSFTPServer):
    def test_init(self):
        # init on a remote url should succeed.
        out, err = self.run_bzr(["init", "--format=pack-0.92", self.get_url()])
        self.assertEqual(out, """Created a standalone branch (format: pack-0.92)\n""")

    def test_init_existing_branch(self):
        # when there is already a branch present, make mention
        self.make_branch(".")

        # rely on SFTPServer get_url() pointing at '.'
        out, err = self.run_bzr_error(["Already a branch"], ["init", self.get_url()])

        # make sure using 'brz checkout' is not suggested
        # for remote locations missing a working tree
        self.assertFalse(re.search(r"use brz checkout", err))

    def test_init_existing_branch_with_workingtree(self):
        # don't distinguish between the branch having a working tree or not
        # when the branch itself is remote.
        self.make_branch_and_tree(".")

        # rely on SFTPServer get_url() pointing at '.'
        self.run_bzr_error(["Already a branch"], ["init", self.get_url()])

    def test_init_append_revisions_only(self):
        self.run_bzr("init --format=dirstate-tags normal_branch6")
        branch = _mod_branch.Branch.open("normal_branch6")
        self.assertEqual(None, branch.get_append_revisions_only())
        self.run_bzr("init --append-revisions-only --format=dirstate-tags branch6")
        branch = _mod_branch.Branch.open("branch6")
        self.assertEqual(True, branch.get_append_revisions_only())
        self.run_bzr_error(
            ["cannot be set to append-revisions-only"],
            "init --append-revisions-only --format=knit knit",
        )

    def test_init_without_username(self):
        """Ensure init works if username is not set."""
        # brz makes user specified whoami mandatory for operations
        # like commit as whoami is recorded. init however is not so final
        # and uses whoami only in a lock file. Without whoami the login name
        # is used. This test is to ensure that init passes even when whoami
        # is not available.
        self.overrideEnv("EMAIL", None)
        self.overrideEnv("BRZ_EMAIL", None)
        out, err = self.run_bzr(["init", "foo"])
        self.assertEqual(err, "")
        self.assertTrue(os.path.exists("foo"))



# Copyright (C) 2005, 2006, 2008, 2009, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for revision properties."""

from breezy.tests import TestNotApplicable
from breezy.tests.per_repository import TestCaseWithRepository


class TestRevProps(TestCaseWithRepository):
    def test_simple_revprops(self):
        """Simple revision properties."""
        wt = self.make_branch_and_tree(".")
        b = wt.branch
        b.nick = "Nicholas"
        if b.repository._format.supports_custom_revision_properties:
            props = {
                "flavor": "choc-mint",
                "condiment": "orange\n  mint\n\tcandy",
                "empty": "",
                "non_ascii": "\xb5",
            }
        else:
            props = {}
        rev1 = wt.commit(
            message="initial null commit", revprops=props, allow_pointless=True
        )
        rev = b.repository.get_revision(rev1)
        if b.repository._format.supports_custom_revision_properties:
            self.assertTrue("flavor" in rev.properties)
            self.assertEqual(rev.properties["flavor"], "choc-mint")
            expected_revprops = {
                "condiment": "orange\n  mint\n\tcandy",
                "empty": "",
                "flavor": "choc-mint",
                "non_ascii": "\xb5",
            }
        else:
            expected_revprops = {}
        if b.repository._format.supports_storing_branch_nick:
            expected_revprops["branch-nick"] = "Nicholas"
        for name, value in expected_revprops.items():
            self.assertEqual(rev.properties[name], value)

    def test_invalid_revprops(self):
        """Invalid revision properties."""
        wt = self.make_branch_and_tree(".")
        b = wt.branch
        if not b.repository._format.supports_custom_revision_properties:
            raise TestNotApplicable(
                "format does not support custom revision properties"
            )
        self.assertRaises(
            ValueError,
            wt.commit,
            message="invalid",
            revprops={"what a silly property": "fine"},
        )
        self.assertRaises(
            ValueError, wt.commit, message="invalid", revprops={"number": 13}
        )


class TestRevisionAttributes(TestCaseWithRepository):
    """Test that revision attributes are correct."""

    def test_revision_accessors(self):
        """Make sure the values that come out of a revision are the
        same as the ones that go in.
        """
        tree1 = self.make_branch_and_tree("br1")
        if tree1.branch.repository._format.supports_custom_revision_properties:
            revprops = {
                "empty": "",
                "value": "one",
                "unicode": "\xb5",
                "multiline": "foo\nbar\n\n",
            }
        else:
            revprops = {}
        # create a revision
        rev1 = tree1.commit(
            message="quux", allow_pointless=True, committer="jaq", revprops=revprops
        )
        self.assertEqual(tree1.branch.last_revision(), rev1)
        rev_a = tree1.branch.repository.get_revision(tree1.branch.last_revision())

        tree2 = self.make_branch_and_tree("br2")
        tree2.commit(
            message=rev_a.message,
            timestamp=rev_a.timestamp,
            timezone=rev_a.timezone,
            committer=rev_a.committer,
            rev_id=(
                rev_a.revision_id
                if tree2.branch.repository._format.supports_setting_revision_ids
                else None
            ),
            revprops=rev_a.properties,
            allow_pointless=True,  # there's nothing in this commit
            strict=True,
            verbose=True,
        )
        rev_b = tree2.branch.repository.get_revision(tree2.branch.last_revision())

        self.assertEqual(rev_a.message, rev_b.message)
        self.assertEqual(rev_a.timestamp, rev_b.timestamp)
        self.assertEqual(rev_a.timezone, rev_b.timezone)
        self.assertEqual(rev_a.committer, rev_b.committer)
        self.assertEqual(rev_a.revision_id, rev_b.revision_id)
        self.assertEqual(rev_a.properties, rev_b.properties)

    def test_zero_timezone(self):
        tree1 = self.make_branch_and_tree("br1")

        # create a revision
        r1 = tree1.commit(message="quux", timezone=0)
        rev_a = tree1.branch.repository.get_revision(r1)
        self.assertEqual(0, rev_a.timezone)



# Copyright (C) 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from unittest import TestCase

import patiencediff

from .. import multiparent, tests

LINES_1 = b"a\nb\nc\nd\ne\n".splitlines(True)
LINES_2 = b"a\nc\nd\ne\n".splitlines(True)
LINES_3 = b"a\nb\nc\nd\n".splitlines(True)
LF_SPLIT_LINES = [b"\x00\n", b"\x00\r\x01\n", b"\x02\r\xff"]


class Mock:
    def __init__(self, **kwargs):
        self.__dict__ = kwargs


class TestMulti(TestCase):
    def test_compare_no_parent(self):
        diff = multiparent.MultiParent.from_lines(LINES_1)
        self.assertEqual([multiparent.NewText(LINES_1)], diff.hunks)

    def test_compare_one_parent(self):
        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2])
        self.assertEqual(
            [
                multiparent.ParentText(0, 0, 0, 1),
                multiparent.NewText([b"b\n"]),
                multiparent.ParentText(0, 1, 2, 3),
            ],
            diff.hunks,
        )

        diff = multiparent.MultiParent.from_lines(LINES_2, [LINES_1])
        self.assertEqual(
            [multiparent.ParentText(0, 0, 0, 1), multiparent.ParentText(0, 2, 1, 3)],
            diff.hunks,
        )

    def test_compare_two_parents(self):
        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
        self.assertEqual(
            [multiparent.ParentText(1, 0, 0, 4), multiparent.ParentText(0, 3, 4, 1)],
            diff.hunks,
        )

    def test_compare_two_parents_blocks(self):
        matcher = patiencediff.PatienceSequenceMatcher(None, LINES_2, LINES_1)
        blocks = matcher.get_matching_blocks()
        diff = multiparent.MultiParent.from_lines(
            LINES_1, [LINES_2, LINES_3], left_blocks=blocks
        )
        self.assertEqual(
            [multiparent.ParentText(1, 0, 0, 4), multiparent.ParentText(0, 3, 4, 1)],
            diff.hunks,
        )

    def test_get_matching_blocks(self):
        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2])
        self.assertEqual(
            [(0, 0, 1), (1, 2, 3), (4, 5, 0)],
            list(diff.get_matching_blocks(0, len(LINES_2))),
        )

        diff = multiparent.MultiParent.from_lines(LINES_2, [LINES_1])
        self.assertEqual(
            [(0, 0, 1), (2, 1, 3), (5, 4, 0)],
            list(diff.get_matching_blocks(0, len(LINES_1))),
        )

    def test_range_iterator(self):
        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
        diff.hunks.append(multiparent.NewText([b"q\n"]))
        self.assertEqual(
            [
                (0, 4, "parent", (1, 0, 4)),
                (4, 5, "parent", (0, 3, 4)),
                (5, 6, "new", [b"q\n"]),
            ],
            list(diff.range_iterator()),
        )

    def test_eq(self):
        diff = multiparent.MultiParent.from_lines(LINES_1)
        diff2 = multiparent.MultiParent.from_lines(LINES_1)
        self.assertEqual(diff, diff2)
        diff3 = multiparent.MultiParent.from_lines(LINES_2)
        self.assertFalse(diff == diff3)
        self.assertFalse(diff == Mock(hunks=[multiparent.NewText(LINES_1)]))
        self.assertEqual(
            multiparent.MultiParent(
                [multiparent.NewText(LINES_1), multiparent.ParentText(0, 1, 2, 3)]
            ),
            multiparent.MultiParent(
                [multiparent.NewText(LINES_1), multiparent.ParentText(0, 1, 2, 3)]
            ),
        )

    def test_to_patch(self):
        self.assertEqual(
            [b"i 1\n", b"a\n", b"\n", b"c 0 1 2 3\n"],
            list(
                multiparent.MultiParent(
                    [multiparent.NewText([b"a\n"]), multiparent.ParentText(0, 1, 2, 3)]
                ).to_patch()
            ),
        )

    def test_from_patch(self):
        self.assertEqual(
            multiparent.MultiParent(
                [multiparent.NewText([b"a\n"]), multiparent.ParentText(0, 1, 2, 3)]
            ),
            multiparent.MultiParent.from_patch(b"i 1\na\n\nc 0 1 2 3"),
        )
        self.assertEqual(
            multiparent.MultiParent(
                [multiparent.NewText([b"a"]), multiparent.ParentText(0, 1, 2, 3)]
            ),
            multiparent.MultiParent.from_patch(b"i 1\na\nc 0 1 2 3\n"),
        )

    def test_binary_content(self):
        patch = list(multiparent.MultiParent.from_lines(LF_SPLIT_LINES).to_patch())
        multiparent.MultiParent.from_patch(b"".join(patch))

    def test_make_patch_from_binary(self):
        patch = multiparent.MultiParent.from_texts(b"".join(LF_SPLIT_LINES))
        expected = multiparent.MultiParent([multiparent.NewText(LF_SPLIT_LINES)])
        self.assertEqual(expected, patch)

    def test_num_lines(self):
        mp = multiparent.MultiParent([multiparent.NewText([b"a\n"])])
        self.assertEqual(1, mp.num_lines())
        mp.hunks.append(multiparent.NewText([b"b\n", b"c\n"]))
        self.assertEqual(3, mp.num_lines())
        mp.hunks.append(multiparent.ParentText(0, 0, 3, 2))
        self.assertEqual(5, mp.num_lines())
        mp.hunks.append(multiparent.NewText([b"f\n", b"g\n"]))
        self.assertEqual(7, mp.num_lines())

    def test_to_lines(self):
        mpdiff = multiparent.MultiParent.from_texts(b"a\nb\nc\n", (b"b\nc\n",))
        lines = mpdiff.to_lines((b"b\ne\n",))
        self.assertEqual([b"a\n", b"b\n", b"e\n"], lines)


class TestNewText(TestCase):
    def test_eq(self):
        self.assertEqual(multiparent.NewText([]), multiparent.NewText([]))
        self.assertFalse(multiparent.NewText([b"a"]) == multiparent.NewText([b"b"]))
        self.assertFalse(multiparent.NewText([b"a"]) == Mock(lines=[b"a"]))

    def test_to_patch(self):
        self.assertEqual([b"i 0\n", b"\n"], list(multiparent.NewText([]).to_patch()))
        self.assertEqual(
            [b"i 1\n", b"a", b"\n"], list(multiparent.NewText([b"a"]).to_patch())
        )
        self.assertEqual(
            [b"i 1\n", b"a\n", b"\n"], list(multiparent.NewText([b"a\n"]).to_patch())
        )


class TestParentText(TestCase):
    def test_eq(self):
        self.assertEqual(
            multiparent.ParentText(1, 2, 3, 4), multiparent.ParentText(1, 2, 3, 4)
        )
        self.assertFalse(
            multiparent.ParentText(1, 2, 3, 4) == multiparent.ParentText(2, 2, 3, 4)
        )
        self.assertFalse(
            multiparent.ParentText(1, 2, 3, 4)
            == Mock(parent=1, parent_pos=2, child_pos=3, num_lines=4)
        )

    def test_to_patch(self):
        self.assertEqual(
            [b"c 0 1 2 3\n"], list(multiparent.ParentText(0, 1, 2, 3).to_patch())
        )


REV_A = [b"a\n", b"b\n", b"c\n", b"d\n"]
REV_B = [b"a\n", b"c\n", b"d\n", b"e\n"]
REV_C = [b"a\n", b"b\n", b"e\n", b"f\n"]


class TestVersionedFile(TestCase):
    def add_version(self, vf, text, version_id, parent_ids):
        vf.add_version(
            [(bytes([t]) + b"\n") for t in bytearray(text)], version_id, parent_ids
        )

    def make_vf(self):
        vf = multiparent.MultiMemoryVersionedFile()
        self.add_version(vf, b"abcd", b"rev-a", [])
        self.add_version(vf, b"acde", b"rev-b", [])
        self.add_version(vf, b"abef", b"rev-c", [b"rev-a", b"rev-b"])
        return vf

    def test_add_version(self):
        vf = self.make_vf()
        self.assertEqual(REV_A, vf._lines[b"rev-a"])
        vf.clear_cache()
        self.assertEqual(vf._lines, {})

    def test_get_line_list(self):
        vf = self.make_vf()
        vf.clear_cache()
        self.assertEqual(REV_A, vf.get_line_list([b"rev-a"])[0])
        self.assertEqual([REV_B, REV_C], vf.get_line_list([b"rev-b", b"rev-c"]))

    def test_reconstruct_empty(self):
        vf = multiparent.MultiMemoryVersionedFile()
        vf.add_version([], b"a", [])
        self.assertEqual([], self.reconstruct_version(vf, b"a"))

    @staticmethod
    def reconstruct(vf, revision_id, start, end):
        reconstructor = multiparent._Reconstructor(vf, vf._lines, vf._parents)
        lines = []
        reconstructor._reconstruct(lines, revision_id, start, end)
        return lines

    @staticmethod
    def reconstruct_version(vf, revision_id):
        reconstructor = multiparent._Reconstructor(vf, vf._lines, vf._parents)
        lines = []
        reconstructor.reconstruct_version(lines, revision_id)
        return lines

    def test_reconstructor(self):
        vf = self.make_vf()
        self.assertEqual([b"a\n", b"b\n"], self.reconstruct(vf, b"rev-a", 0, 2))
        self.assertEqual([b"c\n", b"d\n"], self.reconstruct(vf, b"rev-a", 2, 4))
        self.assertEqual([b"e\n", b"f\n"], self.reconstruct(vf, b"rev-c", 2, 4))
        self.assertEqual(
            [b"a\n", b"b\n", b"e\n", b"f\n"], self.reconstruct(vf, b"rev-c", 0, 4)
        )
        self.assertEqual(
            [b"a\n", b"b\n", b"e\n", b"f\n"], self.reconstruct_version(vf, b"rev-c")
        )

    def test_reordered(self):
        """Check for a corner case that requires re-starting the cursor."""
        vf = multiparent.MultiMemoryVersionedFile()
        # rev-b must have at least two hunks, so split a and b with c.
        self.add_version(vf, b"c", b"rev-a", [])
        self.add_version(vf, b"acb", b"rev-b", [b"rev-a"])
        # rev-c and rev-d must each have a line from a different rev-b hunk
        self.add_version(vf, b"b", b"rev-c", [b"rev-b"])
        self.add_version(vf, b"a", b"rev-d", [b"rev-b"])
        # The lines from rev-c and rev-d must appear in the opposite order
        self.add_version(vf, b"ba", b"rev-e", [b"rev-c", b"rev-d"])
        vf.clear_cache()
        lines = vf.get_line_list([b"rev-e"])[0]
        self.assertEqual([b"b\n", b"a\n"], lines)


class TestMultiVersionedFile(tests.TestCaseInTempDir):
    def test_save_load(self):
        vf = multiparent.MultiVersionedFile("foop")
        vf.add_version(b"a\nb\nc\nd".splitlines(True), b"a", [])
        vf.add_version(b"a\ne\nd\n".splitlines(True), b"b", [b"a"])
        vf.save()
        newvf = multiparent.MultiVersionedFile("foop")
        newvf.load()
        self.assertEqual(b"a\nb\nc\nd", b"".join(newvf.get_line_list([b"a"])[0]))
        self.assertEqual(b"a\ne\nd\n", b"".join(newvf.get_line_list([b"b"])[0]))

    def test_filenames(self):
        vf = multiparent.MultiVersionedFile("foop")
        vf.add_version(b"a\nb\nc\nd".splitlines(True), b"a", [])
        self.assertPathExists("foop.mpknit")
        self.assertPathDoesNotExist("foop.mpidx")
        vf.save()
        self.assertPathExists("foop.mpidx")
        vf.destroy()
        self.assertPathDoesNotExist("foop.mpknit")
        self.assertPathDoesNotExist("foop.mpidx")



# Copyright (C) 2006-2010 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy import (
    branch,
    controldir,
    errors,
    mutabletree,
    osutils,
    tests,
    ui,
)
from breezy import revision as _mod_revision
from breezy.commit import CannotCommitSelectedFileMerge, PointlessCommit
from breezy.tests.matchers import HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree
from breezy.tests.testui import ProgressRecordingUIFactory


class TestCommit(TestCaseWithWorkingTree):
    def test_autodelete_renamed(self):
        tree_a = self.make_branch_and_tree("a")
        self.build_tree(["a/dir/", "a/dir/f1", "a/dir/f2"])
        tree_a.add(["dir", "dir/f1", "dir/f2"])
        tree_a.commit("init")
        # Start off by renaming entries,
        # but then actually auto delete the whole tree
        # https://bugs.launchpad.net/bzr/+bug/114615
        tree_a.rename_one("dir/f1", "dir/a")
        tree_a.rename_one("dir/f2", "dir/z")
        osutils.rmtree("a/dir")
        tree_a.commit("autoremoved")

        with tree_a.lock_read():
            paths = [(path, ie.file_id) for path, ie in tree_a.iter_entries_by_dir()]
        # The only paths left should be the root
        if tree_a.supports_file_ids:
            self.assertEqual([("", tree_a.path2id(""))], paths)

    def test_no_autodelete_renamed_away(self):
        tree_a = self.make_branch_and_tree("a")
        self.build_tree(["a/dir/", "a/dir/f1", "a/dir/f2", "a/dir2/"])
        tree_a.add(["dir", "dir/f1", "dir/f2", "dir2"])
        rev_id1 = tree_a.commit("init")
        revtree = tree_a.branch.repository.revision_tree(rev_id1)
        # Rename one entry out of this directory
        tree_a.rename_one("dir/f1", "dir2/a")
        osutils.rmtree("a/dir")
        tree_a.commit("autoremoved")

        # The only paths left should be the root
        self.assertThat(
            tree_a,
            HasPathRelations(
                revtree, [("", ""), ("dir2/", "dir2/"), ("dir2/a", "dir/f1")]
            ),
        )

    def test_no_autodelete_alternate_renamed(self):
        # Test for bug #114615
        tree_a = self.make_branch_and_tree("A")
        self.build_tree(["A/a/", "A/a/m", "A/a/n"])
        tree_a.add(["a", "a/m", "a/n"])
        tree_a.commit("init")

        tree_b = tree_a.controldir.sprout("B").open_workingtree()
        self.build_tree(["B/xyz/"])
        tree_b.add(["xyz"])
        tree_b.rename_one("a/m", "xyz/m")
        osutils.rmtree("B/a")
        tree_b.commit("delete in B")

        self.assertThat(
            tree_b,
            HasPathRelations(tree_a, [("", ""), ("xyz/", None), ("xyz/m", "a/m")]),
        )

        self.build_tree_contents([("A/a/n", b"new contents for n\n")])
        tree_a.commit("change n in A")

        # Merging from A should introduce conflicts because 'n' was modified
        # (in A) and removed (in B), so 'a' needs to be restored.
        conflicts = tree_b.merge_from_branch(tree_a.branch)
        if tree_b.has_versioned_directories():
            self.assertEqual(3, len(conflicts))
        else:
            self.assertEqual(1, len(conflicts))

        self.assertThat(
            tree_b,
            HasPathRelations(
                tree_a,
                [
                    ("", ""),
                    ("a/", "a/"),
                    ("xyz/", None),
                    ("a/n.OTHER", "a/n"),
                    ("xyz/m", "a/m"),
                ],
            ),
        )

        osutils.rmtree("B/a")
        try:
            # bzr resolve --all
            tree_b.set_conflicts([])
        except errors.UnsupportedOperation:
            # On WT2, set_conflicts is unsupported, but the rmtree has the same
            # effect.
            pass
        tree_b.commit("autoremove a, without touching xyz/m")

        self.assertThat(
            tree_b,
            HasPathRelations(tree_a, [("", ""), ("xyz/", None), ("xyz/m", "a/m")]),
        )

    def test_commit_exclude_pending_merge_fails(self):
        """Excludes are a form of partial commit."""
        wt = self.make_branch_and_tree(".")
        self.build_tree(["foo"])
        wt.add("foo")
        wt.commit("commit one")
        wt2 = wt.controldir.sprout("to").open_workingtree()
        wt2.commit("change_right")
        wt.merge_from_branch(wt2.branch)
        self.assertRaises(
            CannotCommitSelectedFileMerge, wt.commit, "test", exclude=["foo"]
        )

    def test_commit_exclude_exclude_changed_is_pointless(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.smart_add(["."])
        tree.commit("setup test")
        self.build_tree_contents([("a", b'new contents for "a"\n')])
        self.assertRaises(
            PointlessCommit, tree.commit, "test", exclude=["a"], allow_pointless=False
        )

    def test_commit_exclude_excludes_modified_files(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b", "c"])
        tree.smart_add(["."])
        tree.commit("test", exclude=["b", "c"])
        # If b was excluded it will still be 'added' in status.
        tree.lock_read()
        self.addCleanup(tree.unlock)
        changes = list(tree.iter_changes(tree.basis_tree()))
        self.assertEqual([(None, "b"), (None, "c")], [c.path for c in changes])

    def test_commit_exclude_subtree_of_selected(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c"])
        tree.smart_add(["."])
        tree.commit("test", specific_files=["a", "a/c"], exclude=["a/b"])
        # If a/b was excluded it will still be 'added' in status.
        tree.lock_read()
        self.addCleanup(tree.unlock)
        changes = list(tree.iter_changes(tree.basis_tree()))
        self.assertEqual(1, len(changes), changes)
        self.assertEqual((None, "a/b"), changes[0].path)

    def test_commit_sets_last_revision(self):
        tree = self.make_branch_and_tree("tree")
        if tree.branch.repository._format.supports_setting_revision_ids:
            committed_id = tree.commit("foo", rev_id=b"foo")
            # the commit should have returned the same id we asked for.
            self.assertEqual(b"foo", committed_id)
        else:
            committed_id = tree.commit("foo")
        self.assertEqual([committed_id], tree.get_parent_ids())

    def test_commit_returns_revision_id(self):
        tree = self.make_branch_and_tree(".")
        committed_id = tree.commit("message")
        self.assertTrue(tree.branch.repository.has_revision(committed_id))
        self.assertNotEqual(None, committed_id)

    def test_commit_local_unbound(self):
        # using the library api to do a local commit on unbound branches is
        # also an error
        tree = self.make_branch_and_tree("tree")
        self.assertRaises(
            errors.LocalRequiresBoundBranch, tree.commit, "foo", local=True
        )

    def test_commit_merged_kind_change(self):
        """Test merging a kind change.

        Test making a kind change in a working tree, and then merging that
        from another. When committed it should commit the new kind.
        """
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        wt.add(["a"])
        wt.commit("commit one")
        wt2 = wt.controldir.sprout("to").open_workingtree()
        os.remove("a")
        os.mkdir("a")
        wt.commit("changed kind")
        wt2.merge_from_branch(wt.branch)
        wt2.commit("merged kind change")

    def test_commit_aborted_does_not_apply_automatic_changes_bug_282402(self):
        wt = self.make_branch_and_tree(".")
        wt.add(["a"], ["file"])
        self.assertTrue(wt.is_versioned("a"))
        if wt.supports_setting_file_ids():
            a_id = wt.path2id("a")
            self.assertEqual("a", wt.id2path(a_id))

        def fail_message(obj):
            raise errors.CommandError("empty commit message")

        self.assertRaises(errors.CommandError, wt.commit, message_callback=fail_message)
        self.assertTrue(wt.is_versioned("a"))
        if wt.supports_setting_file_ids():
            self.assertEqual("a", wt.id2path(a_id))

    def test_local_commit_ignores_master(self):
        # a --local commit does not require access to the master branch
        # at all, or even for it to exist.
        # we test this by setting up a bound branch and then corrupting
        # the master.
        master = self.make_branch("master")
        tree = self.make_branch_and_tree("tree")
        try:
            tree.branch.bind(master)
        except branch.BindingUnsupported:
            # older format.
            return
        master.controldir.transport.put_bytes("branch-format", b"garbage")
        del master
        # check its corrupted.
        self.assertRaises(
            errors.UnknownFormatError, controldir.ControlDir.open, "master"
        )
        tree.commit("foo", local=True)

    def test_local_commit_does_not_push_to_master(self):
        # a --local commit does not require access to the master branch
        # at all, or even for it to exist.
        # we test that even when its available it does not push to it.
        master = self.make_branch("master")
        tree = self.make_branch_and_tree("tree")
        try:
            tree.branch.bind(master)
        except branch.BindingUnsupported:
            # older format.
            return
        committed_id = tree.commit("foo", local=True)
        self.assertFalse(master.repository.has_revision(committed_id))
        self.assertEqual(_mod_revision.NULL_REVISION, master.last_revision())

    def test_record_initial_ghost(self):
        """The working tree needs to record ghosts during commit."""
        wt = self.make_branch_and_tree(".")
        if not wt.branch.repository._format.supports_ghosts:
            raise tests.TestNotApplicable("format does not support ghosts")
        wt.set_parent_ids(
            [b"non:existent@rev--ision--0--2"], allow_leftmost_as_ghost=True
        )
        rev_id = wt.commit("commit against a ghost first parent.")
        rev = wt.branch.repository.get_revision(rev_id)
        self.assertEqual(rev.parent_ids, [b"non:existent@rev--ision--0--2"])
        # parent_sha1s is not populated now, WTF. rbc 20051003
        self.assertEqual(len(rev.parent_sha1s), 0)

    def test_record_two_ghosts(self):
        """The working tree should preserve all the parents during commit."""
        wt = self.make_branch_and_tree(".")
        if not wt.branch.repository._format.supports_ghosts:
            raise tests.TestNotApplicable("format does not support ghosts")
        wt.set_parent_ids(
            [
                b"foo@azkhazan-123123-abcabc",
                b"wibble@fofof--20050401--1928390812",
            ],
            allow_leftmost_as_ghost=True,
        )
        rev_id = wt.commit("commit from ghost base with one merge")
        # the revision should have been committed with two parents
        rev = wt.branch.repository.get_revision(rev_id)
        self.assertEqual(
            [b"foo@azkhazan-123123-abcabc", b"wibble@fofof--20050401--1928390812"],
            rev.parent_ids,
        )

    def test_commit_deleted_subtree_and_files_updates_workingtree(self):
        """The working trees inventory may be adjusted by commit."""
        wt = self.make_branch_and_tree(".")
        wt.lock_write()
        self.build_tree(["a", "b/", "b/c", "d"])
        wt.add(["a", "b", "b/c", "d"])
        this_dir = wt.controldir.root_transport
        this_dir.delete_tree("b")
        this_dir.delete("d")
        # now we have a tree with a through d in the inventory, but only
        # a present on disk. After commit b-id, c-id and d-id should be
        # missing from the inventory, within the same tree transaction.
        wt.commit("commit stuff")
        self.assertTrue(wt.has_filename("a"))
        self.assertFalse(wt.has_filename("b"))
        self.assertFalse(wt.has_filename("b/c"))
        self.assertFalse(wt.has_filename("d"))
        wt.unlock()
        # the changes should have persisted to disk - reopen the workingtree
        # to be sure.
        wt = wt.controldir.open_workingtree()
        with wt.lock_read():
            self.assertTrue(wt.has_filename("a"))
            self.assertFalse(wt.has_filename("b"))
            self.assertFalse(wt.has_filename("b/c"))
            self.assertFalse(wt.has_filename("d"))

    def test_commit_deleted_subtree_with_removed(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c", "d"])
        wt.add(["a", "b", "b/c"])
        wt.commit("first")
        wt.remove("b/c")
        this_dir = wt.controldir.root_transport
        this_dir.delete_tree("b")
        with wt.lock_write():
            wt.commit("commit deleted rename")
            self.assertTrue(wt.is_versioned("a"))
            self.assertTrue(wt.has_filename("a"))
            self.assertFalse(wt.has_filename("b"))
            self.assertFalse(wt.has_filename("b/c"))

    def test_commit_move_new(self):
        wt = self.make_branch_and_tree("first")
        wt.commit("first")
        wt2 = wt.controldir.sprout("second").open_workingtree()
        self.build_tree(["second/name1"])
        wt2.add("name1")
        wt2.commit("second")
        wt.merge_from_branch(wt2.branch)
        wt.rename_one("name1", "name2")
        wt.commit("third")
        self.assertFalse(wt.is_versioned("name1"))

    def test_nested_commit(self):
        """Commit in multiply-nested trees."""
        tree = self.make_branch_and_tree(".")
        if not tree.supports_tree_reference():
            # inapplicable test.
            return
        subtree = self.make_branch_and_tree("subtree")
        subsubtree = self.make_branch_and_tree("subtree/subtree")
        subsub_revid = subsubtree.commit("subsubtree")
        subtree.commit("subtree")
        subtree.add(["subtree"])
        tree.add(["subtree"])
        # use allow_pointless=False to ensure that the deepest tree, which
        # has no commits made to it, does not get a pointless commit.
        rev_id = tree.commit("added reference", allow_pointless=False)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # the deepest subtree has not changed, so no commit should take place.
        self.assertEqual(subsub_revid, subsubtree.last_revision())
        # the intermediate tree should have committed a pointer to the current
        # subtree revision.
        sub_basis = subtree.basis_tree()
        sub_basis.lock_read()
        self.addCleanup(sub_basis.unlock)
        self.assertEqual(
            subsubtree.last_revision(), sub_basis.get_reference_revision("subtree")
        )
        # the intermediate tree has changed, so should have had a commit
        # take place.
        self.assertNotEqual(None, subtree.last_revision())
        # the outer tree should have committed a pointer to the current
        # subtree revision.
        basis = tree.basis_tree()
        basis.lock_read()
        self.addCleanup(basis.unlock)
        self.assertEqual(
            subtree.last_revision(), basis.get_reference_revision("subtree")
        )
        # the outer tree must have have changed too.
        self.assertNotEqual(None, rev_id)

    def test_nested_commit_second_commit_detects_changes(self):
        """Commit with a nested tree picks up the correct child revid."""
        tree = self.make_branch_and_tree(".")
        if not tree.supports_tree_reference():
            # inapplicable test.
            return
        self.knownFailure("nested trees don't work well with iter_changes")
        subtree = self.make_branch_and_tree("subtree")
        tree.add(["subtree"])
        self.build_tree(["subtree/file"])
        subtree.add(["file"])
        rev_id = tree.commit("added reference", allow_pointless=False)
        tree.get_reference_revision("subtree")
        child_revid = subtree.last_revision()
        # now change the child tree
        self.build_tree_contents([("subtree/file", b"new-content")])
        # and commit in the parent should commit the child and grab its revid,
        # we test with allow_pointless=False here so that we are simulating
        # what users will see.
        rev_id2 = tree.commit("changed subtree only", allow_pointless=False)
        # the child tree has changed, so should have had a commit
        # take place.
        self.assertNotEqual(None, subtree.last_revision())
        self.assertNotEqual(child_revid, subtree.last_revision())
        # the outer tree should have committed a pointer to the current
        # subtree revision.
        basis = tree.basis_tree()
        basis.lock_read()
        self.addCleanup(basis.unlock)
        self.assertEqual(
            subtree.last_revision(), basis.get_reference_revision("subtree")
        )
        self.assertNotEqual(rev_id, rev_id2)

    def test_nested_pointless_commits_are_pointless(self):
        tree = self.make_branch_and_tree(".")
        if not tree.supports_tree_reference():
            # inapplicable test.
            return
        subtree = self.make_branch_and_tree("subtree")
        subtree.commit("")
        tree.add(["subtree"])
        # record the reference.
        rev_id = tree.commit("added reference")
        child_revid = subtree.last_revision()
        # now do a no-op commit with allow_pointless=False
        self.assertRaises(PointlessCommit, tree.commit, "", allow_pointless=False)
        self.assertEqual(child_revid, subtree.last_revision())
        self.assertEqual(rev_id, tree.last_revision())


class TestCommitProgress(TestCaseWithWorkingTree):
    def setUp(self):
        super().setUp()
        ui.ui_factory = ProgressRecordingUIFactory()

    def test_commit_progress_steps(self):
        # during commit we one progress update for every entry in the
        # inventory, and then one for the inventory, and one for the
        # inventory, and one for the revision insertions.
        # first we need a test commit to do. Lets setup a branch with
        # 3 files, and alter one in a selected-file commit. This exercises
        # a number of cases quickly. We should also test things like
        # selective commits which excludes newly added files.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b", "c"])
        tree.add(["a", "b", "c"])
        tree.commit("first post")
        with open("b", "w") as f:
            f.write("new content")
        # set a progress bar that captures the calls so we can see what is
        # emitted
        factory = ProgressRecordingUIFactory()
        ui.ui_factory = factory
        # TODO RBC 20060421 it would be nice to merge the reporter output
        # into the factory for this test - just make the test ui factory
        # pun as a reporter. Then we can check the ordering is right.
        tree.commit("second post", specific_files=["b"])
        # 5 steps, the first of which is reported 2 times, once per dir
        self.assertEqual(
            [
                ("update", 1, 5, "Collecting changes [0] - Stage"),
                ("update", 1, 5, "Collecting changes [1] - Stage"),
                ("update", 2, 5, "Saving data locally - Stage"),
                ("update", 3, 5, "Running pre_commit hooks - Stage"),
                ("update", 4, 5, "Updating the working tree - Stage"),
                ("update", 5, 5, "Running post_commit hooks - Stage"),
            ],
            factory._calls,
        )

    def test_commit_progress_shows_post_hook_names(self):
        tree = self.make_branch_and_tree(".")
        # set a progress bar that captures the calls so we can see what is
        # emitted
        factory = ProgressRecordingUIFactory()
        ui.ui_factory = factory

        def a_hook(_, _2, _3, _4, _5, _6):
            pass

        branch.Branch.hooks.install_named_hook("post_commit", a_hook, "hook name")
        tree.commit("first post")
        self.assertEqual(
            [
                ("update", 1, 5, "Collecting changes [0] - Stage"),
                ("update", 1, 5, "Collecting changes [1] - Stage"),
                ("update", 2, 5, "Saving data locally - Stage"),
                ("update", 3, 5, "Running pre_commit hooks - Stage"),
                ("update", 4, 5, "Updating the working tree - Stage"),
                ("update", 5, 5, "Running post_commit hooks - Stage"),
                ("update", 5, 5, "Running post_commit hooks [hook name] - Stage"),
            ],
            factory._calls,
        )

    def test_commit_progress_shows_pre_hook_names(self):
        tree = self.make_branch_and_tree(".")
        # set a progress bar that captures the calls so we can see what is
        # emitted
        factory = ProgressRecordingUIFactory()
        ui.ui_factory = factory

        def a_hook(_, _2, _3, _4, _5, _6, _7, _8):
            pass

        branch.Branch.hooks.install_named_hook("pre_commit", a_hook, "hook name")
        tree.commit("first post")
        self.assertEqual(
            [
                ("update", 1, 5, "Collecting changes [0] - Stage"),
                ("update", 1, 5, "Collecting changes [1] - Stage"),
                ("update", 2, 5, "Saving data locally - Stage"),
                ("update", 3, 5, "Running pre_commit hooks - Stage"),
                ("update", 3, 5, "Running pre_commit hooks [hook name] - Stage"),
                ("update", 4, 5, "Updating the working tree - Stage"),
                ("update", 5, 5, "Running post_commit hooks - Stage"),
            ],
            factory._calls,
        )

    def test_start_commit_hook(self):
        """Make sure a start commit hook can modify the tree that is
        committed.
        """

        def start_commit_hook_adds_file(tree):
            with open(tree.abspath("newfile"), "w") as f:
                f.write("data")
            tree.add(["newfile"])

        def restoreDefaults():
            mutabletree.MutableTree.hooks["start_commit"] = []

        self.addCleanup(restoreDefaults)
        tree = self.make_branch_and_tree(".")
        mutabletree.MutableTree.hooks.install_named_hook(
            "start_commit", start_commit_hook_adds_file, None
        )
        tree.commit("first post")
        committed_tree = tree.basis_tree()
        self.assertTrue(committed_tree.has_filename("newfile"))

    def test_post_commit_hook(self):
        """Make sure a post_commit hook is called after a commit."""

        def post_commit_hook_test_params(params):
            self.assertTrue(isinstance(params, mutabletree.PostCommitHookParams))
            self.assertTrue(isinstance(params.mutable_tree, mutabletree.MutableTree))
            with open(tree.abspath("newfile"), "w") as f:
                f.write("data")
            params.mutable_tree.add(["newfile"])

        tree = self.make_branch_and_tree(".")
        mutabletree.MutableTree.hooks.install_named_hook(
            "post_commit", post_commit_hook_test_params, None
        )
        self.assertFalse(tree.has_filename("newfile"))
        tree.commit("first post")
        self.assertTrue(tree.has_filename("newfile"))
        committed_tree = tree.basis_tree()
        self.assertFalse(committed_tree.has_filename("newfile"))



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from .. import cmdline, tests
from .features import backslashdir_feature


class TestSplitter(tests.TestCase):
    def assertAsTokens(self, expected, line, single_quotes_allowed=False):
        s = cmdline.Splitter(line, single_quotes_allowed=single_quotes_allowed)
        self.assertEqual(expected, list(s))

    def test_simple(self):
        self.assertAsTokens(
            [(False, "foo"), (False, "bar"), (False, "baz")], "foo bar baz"
        )

    def test_ignore_multiple_spaces(self):
        self.assertAsTokens([(False, "foo"), (False, "bar")], "foo  bar")

    def test_ignore_leading_space(self):
        self.assertAsTokens([(False, "foo"), (False, "bar")], "  foo bar")

    def test_ignore_trailing_space(self):
        self.assertAsTokens([(False, "foo"), (False, "bar")], "foo bar  ")

    def test_posix_quotations(self):
        self.assertAsTokens(
            [(True, "foo bar")], "'foo bar'", single_quotes_allowed=True
        )
        self.assertAsTokens(
            [(True, "foo bar")], "'fo''o b''ar'", single_quotes_allowed=True
        )
        self.assertAsTokens(
            [(True, "foo bar")], '"fo""o b""ar"', single_quotes_allowed=True
        )
        self.assertAsTokens(
            [(True, "foo bar")], '"fo"\'o b\'"ar"', single_quotes_allowed=True
        )

    def test_nested_quotations(self):
        self.assertAsTokens([(True, 'foo"" bar')], '"foo\\"\\" bar"')
        self.assertAsTokens([(True, "foo'' bar")], "\"foo'' bar\"")
        self.assertAsTokens(
            [(True, "foo'' bar")], "\"foo'' bar\"", single_quotes_allowed=True
        )
        self.assertAsTokens(
            [(True, 'foo"" bar')], "'foo\"\" bar'", single_quotes_allowed=True
        )

    def test_empty_result(self):
        self.assertAsTokens([], "")
        self.assertAsTokens([], "    ")

    def test_quoted_empty(self):
        self.assertAsTokens([(True, "")], '""')
        self.assertAsTokens([(False, "''")], "''")
        self.assertAsTokens([(True, "")], "''", single_quotes_allowed=True)
        self.assertAsTokens([(False, "a"), (True, ""), (False, "c")], 'a "" c')
        self.assertAsTokens(
            [(False, "a"), (True, ""), (False, "c")],
            "a '' c",
            single_quotes_allowed=True,
        )

    def test_unicode_chars(self):
        self.assertAsTokens(
            [(False, "f\xb5\xee"), (False, "\u1234\u3456")], "f\xb5\xee \u1234\u3456"
        )

    def test_newline_in_quoted_section(self):
        self.assertAsTokens([(True, "foo\nbar\nbaz\n")], '"foo\nbar\nbaz\n"')
        self.assertAsTokens(
            [(True, "foo\nbar\nbaz\n")], "'foo\nbar\nbaz\n'", single_quotes_allowed=True
        )

    def test_escape_chars(self):
        self.assertAsTokens([(False, "foo\\bar")], "foo\\bar")

    def test_escape_quote(self):
        self.assertAsTokens([(True, 'foo"bar')], '"foo\\"bar"')
        self.assertAsTokens([(True, 'foo\\"bar')], '"foo\\\\\\"bar"')
        self.assertAsTokens([(True, "foo\\bar")], '"foo\\\\"bar"')

    def test_double_escape(self):
        self.assertAsTokens([(True, "foo\\\\bar")], '"foo\\\\bar"')
        self.assertAsTokens([(False, "foo\\\\bar")], "foo\\\\bar")

    def test_multiple_quoted_args(self):
        self.assertAsTokens([(True, "x x"), (True, "y y")], '"x x" "y y"')
        self.assertAsTokens(
            [(True, "x x"), (True, "y y")], "\"x x\" 'y y'", single_quotes_allowed=True
        )

    def test_n_backslashes_handling(self):
        # https://bugs.launchpad.net/bzr/+bug/528944
        # actually we care about the doubled backslashes when they're
        # represents UNC paths.
        # But in fact there is too much weird corner cases
        # (see https://bugs.launchpad.net/tortoisebzr/+bug/569050)
        # so to reproduce every bit of windows command-line handling
        # could be not worth of efforts?
        self.requireFeature(backslashdir_feature)
        self.assertAsTokens([(True, r"\\host\path")], r'"\\host\path"')
        self.assertAsTokens([(False, r"\\host\path")], r"\\host\path")
        # handling of " after the 2n and 2n+1 backslashes
        # inside and outside the quoted string
        self.assertAsTokens([(True, r"\\"), (False, r"*.py")], r'"\\\\" *.py')
        self.assertAsTokens([(True, r'\\" *.py')], r'"\\\\\" *.py"')
        self.assertAsTokens([(True, r"\\ *.py")], r'\\\\" *.py"')
        self.assertAsTokens([(False, r'\\"'), (False, r"*.py")], r"\\\\\" *.py")
        self.assertAsTokens([(True, "\\\\")], '"\\\\')



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Infrastructure for server-side request handlers.

Interesting module attributes:
    * The request_handlers registry maps verb names to SmartServerRequest
      classes.
    * The jail_info threading.local() object is used to prevent accidental
      opening of BzrDirs outside of the backing transport, or any other
      transports placed in jail_info.transports.  The jail_info is reset on
      every call into a request handler (which can happen an arbitrary number
      of times during a request).
"""

# XXX: The class names are a little confusing: the protocol will instantiate a
# SmartServerRequestHandler, whose dispatch_command method creates an instance
# of a SmartServerRequest subclass.

import threading
from _thread import get_ident

from ... import branch as _mod_branch
from ... import debug, errors, osutils, registry, revision, trace, urlutils
from ... import transport as _mod_transport
from ...lazy_import import lazy_import

lazy_import(
    globals(),
    """
from breezy.bzr import bzrdir
from breezy.bzr.bundle import serializer

import tempfile
""",
)


jail_info = threading.local()
jail_info.transports = None


class DisabledMethod(errors.InternalBzrError):
    _fmt = "The smart server method '%(class_name)s' is disabled."

    def __init__(self, class_name):
        errors.BzrError.__init__(self)
        self.class_name = class_name


def _install_hook():
    bzrdir.BzrDir.hooks.install_named_hook(
        "pre_open", _pre_open_hook, "checking server jail"
    )


def _pre_open_hook(transport):
    allowed_transports = getattr(jail_info, "transports", None)
    if allowed_transports is None:
        return
    abspath = transport.base
    for allowed_transport in allowed_transports:
        try:
            allowed_transport.relpath(abspath)
        except errors.PathNotChild:
            continue
        else:
            return
    raise errors.JailBreak(abspath)


_install_hook()


class SmartServerRequest:
    """Base class for request handlers.

    To define a new request, subclass this class and override the `do` method
    (and if appropriate, `do_body` as well).  Request implementors should take
    care to call `translate_client_path` and `transport_from_client_path` as
    appropriate when dealing with paths received from the client.
    """

    # XXX: rename this class to BaseSmartServerRequestHandler ?  A request
    # *handler* is a different concept to the request.

    def __init__(self, backing_transport, root_client_path="/", jail_root=None):
        """Constructor.

        :param backing_transport: the base transport to be used when performing
            this request.
        :param root_client_path: the client path that maps to the root of
            backing_transport.  This is used to interpret relpaths received
            from the client.  Clients will not be able to refer to paths above
            this root.  If root_client_path is None, then no translation will
            be performed on client paths.  Default is '/'.
        :param jail_root: if specified, the root of the BzrDir.open jail to use
            instead of backing_transport.
        """
        self._backing_transport = backing_transport
        if jail_root is None:
            jail_root = backing_transport
        self._jail_root = jail_root
        if root_client_path is not None:
            if not root_client_path.startswith("/"):
                root_client_path = "/" + root_client_path
            if not root_client_path.endswith("/"):
                root_client_path += "/"
        self._root_client_path = root_client_path
        self._body_chunks = []

    def _check_enabled(self):
        """Raises DisabledMethod if this method is disabled."""
        pass

    def do(self, *args):
        """Mandatory extension point for SmartServerRequest subclasses.

        Subclasses must implement this.

        This should return a SmartServerResponse if this command expects to
        receive no body.
        """
        raise NotImplementedError(self.do)

    def execute(self, *args):
        """Public entry point to execute this request.

        It will return a SmartServerResponse if the command does not expect a
        body.

        :param args: the arguments of the request.
        """
        self._check_enabled()
        return self.do(*args)

    def do_body(self, body_bytes):
        """Called if the client sends a body with the request.

        The do() method is still called, and must have returned None.

        Must return a SmartServerResponse.
        """
        if body_bytes != b"":
            raise errors.SmartProtocolError("Request does not expect a body")

    def do_chunk(self, chunk_bytes):
        """Called with each body chunk if the request has a streamed body.

        The do() method is still called, and must have returned None.
        """
        self._body_chunks.append(chunk_bytes)

    def do_end(self):
        """Called when the end of the request has been received."""
        body_bytes = b"".join(self._body_chunks)
        self._body_chunks = None
        return self.do_body(body_bytes)

    def setup_jail(self):
        jail_info.transports = [self._jail_root]

    def teardown_jail(self):
        jail_info.transports = None

    def translate_client_path(self, client_path):
        """Translate a path received from a network client into a local
        relpath.

        All paths received from the client *must* be translated.

        :param client_path: the path from the client.
        :returns: a relpath that may be used with self._backing_transport
            (unlike the untranslated client_path, which must not be used with
            the backing transport).
        """
        client_path = client_path.decode("utf-8")
        if self._root_client_path is None:
            # no translation necessary!
            return client_path
        if not client_path.startswith("/"):
            client_path = "/" + client_path
        if client_path + "/" == self._root_client_path:
            return "."
        if client_path.startswith(self._root_client_path):
            path = client_path[len(self._root_client_path) :]
            relpath = urlutils.joinpath("/", path)
            if not relpath.startswith("/"):
                raise ValueError(relpath)
            return urlutils.escape("." + relpath)
        else:
            raise errors.PathNotChild(client_path, self._root_client_path)

    def transport_from_client_path(self, client_path):
        """Get a backing transport corresponding to the location referred to by
        a network client.

        :seealso: translate_client_path
        :returns: a transport cloned from self._backing_transport
        """
        relpath = self.translate_client_path(client_path)
        return self._backing_transport.clone(relpath)


class SmartServerResponse:
    """A response to a client request.

    This base class should not be used. Instead use
    SuccessfulSmartServerResponse and FailedSmartServerResponse as appropriate.
    """

    def __init__(self, args, body=None, body_stream=None):
        """Constructor.

        :param args: tuple of response arguments.
        :param body: string of a response body.
        :param body_stream: iterable of bytestrings to be streamed to the
            client.
        """
        self.args = args
        if body is not None and body_stream is not None:
            raise errors.BzrError("'body' and 'body_stream' are mutually exclusive.")
        self.body = body
        self.body_stream = body_stream

    def __eq__(self, other):
        if other is None:
            return False
        return (
            other.args == self.args
            and other.body == self.body
            and other.body_stream is self.body_stream
        )

    def __repr__(self):
        return "<{} args={!r} body={!r}>".format(
            self.__class__.__name__, self.args, self.body
        )


class FailedSmartServerResponse(SmartServerResponse):
    """A SmartServerResponse for a request which failed."""

    def is_successful(self):
        """FailedSmartServerResponse are not successful."""
        return False


class SuccessfulSmartServerResponse(SmartServerResponse):
    """A SmartServerResponse for a successfully completed request."""

    def is_successful(self):
        """SuccessfulSmartServerResponse are successful."""
        return True


class SmartServerRequestHandler:
    """Protocol logic for smart server.

    This doesn't handle serialization at all, it just processes requests and
    creates responses.
    """

    # IMPORTANT FOR IMPLEMENTORS: It is important that SmartServerRequestHandler
    # not contain encoding or decoding logic to allow the wire protocol to vary
    # from the object protocol: we will want to tweak the wire protocol separate
    # from the object model, and ideally we will be able to do that without
    # having a SmartServerRequestHandler subclass for each wire protocol, rather
    # just a Protocol subclass.

    # TODO: Better way of representing the body for commands that take it,
    # and allow it to be streamed into the server.

    def __init__(self, backing_transport, commands, root_client_path, jail_root=None):
        """Constructor.

        :param backing_transport: a Transport to handle requests for.
        :param commands: a registry mapping command names to SmartServerRequest
            subclasses. e.g. breezy.transport.smart.vfs.vfs_commands.
        """
        self._backing_transport = backing_transport
        self._root_client_path = root_client_path
        self._commands = commands
        if jail_root is None:
            jail_root = backing_transport
        self._jail_root = jail_root
        self.response = None
        self.finished_reading = False
        self._command = None
        if "hpss" in debug.debug_flags:
            self._request_start_time = osutils.perf_counter()
            self._thread_id = get_ident()

    def _trace(self, action, message, extra_bytes=None, include_time=False):
        # It is a bit of a shame that this functionality overlaps with that of
        # ProtocolThreeRequester._trace. However, there is enough difference
        # that just putting it in a helper doesn't help a lot. And some state
        # is taken from the instance.
        if include_time:
            t = "%5.3fs " % (osutils.perf_counter() - self._request_start_time)
        else:
            t = ""
        if extra_bytes is None:
            extra = ""
        else:
            extra = " " + repr(extra_bytes[:40])
            if len(extra) > 33:
                extra = extra[:29] + extra[-1] + "..."
        trace.mutter("%12s: [%s] %s%s%s", action, self._thread_id, t, message, extra)

    def accept_body(self, bytes):
        """Accept body data."""
        if self._command is None:
            # no active command object, so ignore the event.
            return
        self._run_handler_code(self._command.do_chunk, (bytes,), {})
        if "hpss" in debug.debug_flags:
            self._trace("accept body", f"{len(bytes)} bytes", bytes)

    def end_of_body(self):
        """No more body data will be received."""
        self._run_handler_code(self._command.do_end, (), {})
        # cannot read after this.
        self.finished_reading = True
        if "hpss" in debug.debug_flags:
            self._trace("end of body", "", include_time=True)

    def _run_handler_code(self, callable, args, kwargs):
        """Run some handler specific code 'callable'.

        If a result is returned, it is considered to be the commands response,
        and finished_reading is set true, and its assigned to self.response.

        Any exceptions caught are translated and a response object created
        from them.
        """
        result = self._call_converting_errors(callable, args, kwargs)

        if result is not None:
            self.response = result
            self.finished_reading = True

    def _call_converting_errors(self, callable, args, kwargs):
        """Call callable converting errors to Response objects."""
        # XXX: most of this error conversion is VFS-related, and thus ought to
        # be in SmartServerVFSRequestHandler somewhere.
        try:
            self._command.setup_jail()
            try:
                return callable(*args, **kwargs)
            finally:
                self._command.teardown_jail()
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as err:
            err_struct = _translate_error(err)
            return FailedSmartServerResponse(err_struct)

    def headers_received(self, headers):
        # Just a no-op at the moment.
        if "hpss" in debug.debug_flags:
            self._trace("headers", repr(headers))

    def args_received(self, args):
        cmd = args[0]
        args = args[1:]
        try:
            command = self._commands.get(cmd)
        except LookupError as e:
            if "hpss" in debug.debug_flags:
                self._trace("hpss unknown request", cmd, repr(args)[1:-1])
            raise errors.UnknownSmartMethod(cmd) from e
        if "hpss" in debug.debug_flags:
            from . import vfs

            if issubclass(command, vfs.VfsRequest):
                action = "hpss vfs req"
            else:
                action = "hpss request"
            self._trace(action, "{} {}".format(cmd, repr(args)[1:-1]))
        self._command = command(
            self._backing_transport, self._root_client_path, self._jail_root
        )
        self._run_handler_code(self._command.execute, args, {})

    def end_received(self):
        if self._command is None:
            # no active command object, so ignore the event.
            return
        self._run_handler_code(self._command.do_end, (), {})
        if "hpss" in debug.debug_flags:
            self._trace("end", "", include_time=True)

    def post_body_error_received(self, error_args):
        # Just a no-op at the moment.
        pass


def _translate_error(err):
    if isinstance(err, _mod_transport.NoSuchFile):
        return (b"NoSuchFile", err.path.encode("utf-8"))
    elif isinstance(err, _mod_transport.FileExists):
        return (b"FileExists", err.path.encode("utf-8"))
    elif isinstance(err, errors.DirectoryNotEmpty):
        return (b"DirectoryNotEmpty", err.path.encode("utf-8"))
    elif isinstance(err, errors.IncompatibleRepositories):
        return (
            b"IncompatibleRepositories",
            str(err.source),
            str(err.target),
            str(err.details),
        )
    elif isinstance(err, errors.ShortReadvError):
        return (
            b"ShortReadvError",
            err.path.encode("utf-8"),
            str(err.offset).encode("ascii"),
            str(err.length).encode("ascii"),
            str(err.actual).encode("ascii"),
        )
    elif isinstance(err, errors.RevisionNotPresent):
        return (b"RevisionNotPresent", err.revision_id, err.file_id)
    elif isinstance(err, errors.UnstackableRepositoryFormat):
        return (
            b"UnstackableRepositoryFormat",
            str(err.format).encode("utf-8"),
            err.url.encode("utf-8"),
        )
    elif isinstance(err, _mod_branch.UnstackableBranchFormat):
        return (
            b"UnstackableBranchFormat",
            str(err.format).encode("utf-8"),
            err.url.encode("utf-8"),
        )
    elif isinstance(err, errors.NotStacked):
        return (b"NotStacked",)
    elif isinstance(err, errors.BzrCheckError):
        return (b"BzrCheckError", err.msg.encode("utf-8"))
    elif isinstance(err, UnicodeError):
        # If it is a DecodeError, than most likely we are starting
        # with a plain string
        str_or_unicode = err.object
        if isinstance(str_or_unicode, str):
            # XXX: UTF-8 might have \x01 (our protocol v1 and v2 seperator
            # byte) in it, so this encoding could cause broken responses.
            # Newer clients use protocol v3, so will be fine.
            val = "u:" + str_or_unicode.encode("utf-8")
        else:
            val = "s:" + str_or_unicode.encode("base64")
        # This handles UnicodeEncodeError or UnicodeDecodeError
        return (
            err.__class__.__name__,
            err.encoding,
            val,
            str(err.start),
            str(err.end),
            err.reason,
        )
    elif isinstance(err, errors.TransportNotPossible):
        if err.msg == "readonly transport":
            return (b"ReadOnlyError",)
    elif isinstance(err, errors.ReadError):
        # cannot read the file
        return (b"ReadError", err.path)
    elif isinstance(err, errors.PermissionDenied):
        return (
            b"PermissionDenied",
            err.path.encode("utf-8"),
            err.extra.encode("utf-8"),
        )
    elif isinstance(err, errors.TokenMismatch):
        return (b"TokenMismatch", err.given_token, err.lock_token)
    elif isinstance(err, errors.LockContention):
        return (b"LockContention",)
    elif isinstance(err, errors.GhostRevisionsHaveNoRevno):
        return (b"GhostRevisionsHaveNoRevno", err.revision_id, err.ghost_revision_id)
    elif isinstance(err, urlutils.InvalidURL):
        return (b"InvalidURL", err.path.encode("utf-8"), err.extra.encode("ascii"))
    elif isinstance(err, MemoryError):
        # GZ 2011-02-24: Copy breezy.trace -Dmem_dump functionality here?
        return (b"MemoryError",)
    elif isinstance(err, errors.AlreadyControlDirError):
        return (b"AlreadyControlDir", err.path)
    # Unserialisable error.  Log it, and return a generic error
    trace.log_exception_quietly()
    return (
        b"error",
        trace._qualified_exception_name(err.__class__, True).encode("utf-8"),
        str(err).encode("utf-8"),
    )


class HelloRequest(SmartServerRequest):
    """Answer a version request with the highest protocol version this server
    supports.
    """

    def do(self):
        return SuccessfulSmartServerResponse((b"ok", b"2"))


class GetBundleRequest(SmartServerRequest):
    """Get a bundle of from the null revision to the specified revision."""

    def do(self, path, revision_id):
        # open transport relative to our base
        t = self.transport_from_client_path(path)
        control, extra_path = bzrdir.BzrDir.open_containing_from_transport(t)
        repo = control.open_repository()
        tmpf = tempfile.TemporaryFile()
        base_revision = revision.NULL_REVISION
        serializer.write_bundle(repo, revision_id, base_revision, tmpf)
        tmpf.seek(0)
        return SuccessfulSmartServerResponse((), tmpf.read())


class SmartServerIsReadonly(SmartServerRequest):
    # XXX: this request method belongs somewhere else.

    def do(self):
        if self._backing_transport.is_readonly():
            answer = b"yes"
        else:
            answer = b"no"
        return SuccessfulSmartServerResponse((answer,))


# In the 'info' attribute, we store whether this request is 'safe' to retry if
# we get a disconnect while reading the response. It can have the values:
#   read    This is purely a read request, so retrying it is perfectly ok.
#   idem    An idempotent write request. Something like 'put' where if you put
#           the same bytes twice you end up with the same final bytes.
#   semi    This is a request that isn't strictly idempotent, but doesn't
#           result in corruption if it is retried. This is for things like
#           'lock' and 'unlock'. If you call lock, it updates the disk
#           structure. If you fail to read the response, you won't be able to
#           use the lock, because you don't have the lock token. Calling lock
#           again will fail, because the lock is already taken. However, we
#           can't tell if the server received our request or not. If it didn't,
#           then retrying the request is fine, as it will actually do what we
#           want. If it did, we will interrupt the current operation, but we
#           are no worse off than interrupting the current operation because of
#           a ConnectionReset.
#   semivfs Similar to semi, but specific to a Virtual FileSystem request.
#   stream  This is a request that takes a stream that cannot be restarted if
#           consumed. This request is 'safe' in that if we determine the
#           connection is closed before we consume the stream, we can try
#           again.
#   mutate  State is updated in a way that replaying that request results in a
#           different state. For example 'append' writes more bytes to a given
#           file. If append succeeds, it moves the file pointer.
request_handlers = registry.Registry[bytes, SmartServerRequest]()
request_handlers.register_lazy(
    b"append", "breezy.bzr.smart.vfs", "AppendRequest", info="mutate"
)
request_handlers.register_lazy(
    b"Branch.break_lock",
    "breezy.bzr.smart.branch",
    "SmartServerBranchBreakLock",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.get_config_file",
    "breezy.bzr.smart.branch",
    "SmartServerBranchGetConfigFile",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.get_parent",
    "breezy.bzr.smart.branch",
    "SmartServerBranchGetParent",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.put_config_file",
    "breezy.bzr.smart.branch",
    "SmartServerBranchPutConfigFile",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.get_tags_bytes",
    "breezy.bzr.smart.branch",
    "SmartServerBranchGetTagsBytes",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.set_tags_bytes",
    "breezy.bzr.smart.branch",
    "SmartServerBranchSetTagsBytes",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.heads_to_fetch",
    "breezy.bzr.smart.branch",
    "SmartServerBranchHeadsToFetch",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.get_stacked_on_url",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestGetStackedOnURL",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.get_physical_lock_status",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestGetPhysicalLockStatus",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.last_revision_info",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestLastRevisionInfo",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.lock_write",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestLockWrite",
    info="semi",
)
request_handlers.register_lazy(
    b"Branch.revision_history",
    "breezy.bzr.smart.branch",
    "SmartServerRequestRevisionHistory",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.set_config_option",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetConfigOption",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.set_config_option_dict",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetConfigOptionDict",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.set_last_revision",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetLastRevision",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.set_last_revision_info",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetLastRevisionInfo",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.set_last_revision_ex",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetLastRevisionEx",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.set_parent_location",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestSetParentLocation",
    info="idem",
)
request_handlers.register_lazy(
    b"Branch.unlock",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestUnlock",
    info="semi",
)
request_handlers.register_lazy(
    b"Branch.revision_id_to_revno",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestRevisionIdToRevno",
    info="read",
)
request_handlers.register_lazy(
    b"Branch.get_all_reference_info",
    "breezy.bzr.smart.branch",
    "SmartServerBranchRequestGetAllReferenceInfo",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.checkout_metadir",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestCheckoutMetaDir",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.cloning_metadir",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestCloningMetaDir",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.create_branch",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestCreateBranch",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDir.create_repository",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestCreateRepository",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDir.find_repository",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestFindRepositoryV1",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.find_repositoryV2",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestFindRepositoryV2",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.find_repositoryV3",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestFindRepositoryV3",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.get_branches",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestGetBranches",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.get_config_file",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestConfigFile",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.destroy_branch",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestDestroyBranch",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDir.destroy_repository",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestDestroyRepository",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDir.has_workingtree",
    "breezy.bzr.smart.bzrdir",
    "SmartServerBzrDirRequestHasWorkingTree",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDirFormat.initialize",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestInitializeBzrDir",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDirFormat.initialize_ex_1.16",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestBzrDirInitializeEx",
    info="semi",
)
request_handlers.register_lazy(
    b"BzrDir.open",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestOpenBzrDir",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.open_2.1",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestOpenBzrDir_2_1",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.open_branch",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestOpenBranch",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.open_branchV2",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestOpenBranchV2",
    info="read",
)
request_handlers.register_lazy(
    b"BzrDir.open_branchV3",
    "breezy.bzr.smart.bzrdir",
    "SmartServerRequestOpenBranchV3",
    info="read",
)
request_handlers.register_lazy(
    b"delete", "breezy.bzr.smart.vfs", "DeleteRequest", info="semivfs"
)
request_handlers.register_lazy(
    b"get", "breezy.bzr.smart.vfs", "GetRequest", info="read"
)
request_handlers.register_lazy(
    b"get_bundle", "breezy.bzr.smart.request", "GetBundleRequest", info="read"
)
request_handlers.register_lazy(
    b"has", "breezy.bzr.smart.vfs", "HasRequest", info="read"
)
request_handlers.register_lazy(
    b"hello", "breezy.bzr.smart.request", "HelloRequest", info="read"
)
request_handlers.register_lazy(
    b"iter_files_recursive",
    "breezy.bzr.smart.vfs",
    "IterFilesRecursiveRequest",
    info="read",
)
request_handlers.register_lazy(
    b"list_dir", "breezy.bzr.smart.vfs", "ListDirRequest", info="read"
)
request_handlers.register_lazy(
    b"mkdir", "breezy.bzr.smart.vfs", "MkdirRequest", info="semivfs"
)
request_handlers.register_lazy(
    b"move", "breezy.bzr.smart.vfs", "MoveRequest", info="semivfs"
)
request_handlers.register_lazy(
    b"put", "breezy.bzr.smart.vfs", "PutRequest", info="idem"
)
request_handlers.register_lazy(
    b"put_non_atomic", "breezy.bzr.smart.vfs", "PutNonAtomicRequest", info="idem"
)
request_handlers.register_lazy(
    b"readv", "breezy.bzr.smart.vfs", "ReadvRequest", info="read"
)
request_handlers.register_lazy(
    b"rename", "breezy.bzr.smart.vfs", "RenameRequest", info="semivfs"
)
request_handlers.register_lazy(
    b"Repository.add_signature_text",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryAddSignatureText",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.annotate_file_revision",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryAnnotateFileRevision",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.all_revision_ids",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryAllRevisionIds",
    info="read",
)
request_handlers.register_lazy(
    b"PackRepository.autopack",
    "breezy.bzr.smart.packrepository",
    "SmartServerPackRepositoryAutopack",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.break_lock",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryBreakLock",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.gather_stats",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGatherStats",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_parent_map",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetParentMap",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_revision_graph",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetRevisionGraph",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_revision_signature_text",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetRevisionSignatureText",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.has_revision",
    "breezy.bzr.smart.repository",
    "SmartServerRequestHasRevision",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.has_signature_for_revision_id",
    "breezy.bzr.smart.repository",
    "SmartServerRequestHasSignatureForRevisionId",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.insert_stream",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryInsertStream",
    info="stream",
)
request_handlers.register_lazy(
    b"Repository.insert_stream_1.19",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryInsertStream_1_19",
    info="stream",
)
request_handlers.register_lazy(
    b"Repository.insert_stream_locked",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryInsertStreamLocked",
    info="stream",
)
request_handlers.register_lazy(
    b"Repository.is_shared",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryIsShared",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.iter_files_bytes",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryIterFilesBytes",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.lock_write",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryLockWrite",
    info="semi",
)
request_handlers.register_lazy(
    b"Repository.make_working_trees",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryMakeWorkingTrees",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.set_make_working_trees",
    "breezy.bzr.smart.repository",
    "SmartServerRepositorySetMakeWorkingTrees",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.unlock",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryUnlock",
    info="semi",
)
request_handlers.register_lazy(
    b"Repository.get_physical_lock_status",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetPhysicalLockStatus",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_rev_id_for_revno",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetRevIdForRevno",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_stream",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetStream",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_stream_1.19",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetStream_1_19",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.get_stream_for_missing_keys",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetStreamForMissingKeys",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.iter_revisions",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryIterRevisions",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.pack",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryPack",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.start_write_group",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryStartWriteGroup",
    info="semi",
)
request_handlers.register_lazy(
    b"Repository.commit_write_group",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryCommitWriteGroup",
    info="semi",
)
request_handlers.register_lazy(
    b"Repository.abort_write_group",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryAbortWriteGroup",
    info="semi",
)
request_handlers.register_lazy(
    b"Repository.check_write_group",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryCheckWriteGroup",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.reconcile",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryReconcile",
    info="idem",
)
request_handlers.register_lazy(
    b"Repository.revision_archive",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryRevisionArchive",
    info="read",
)
request_handlers.register_lazy(
    b"Repository.tarball",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryTarball",
    info="read",
)
request_handlers.register_lazy(
    b"VersionedFileRepository.get_serializer_format",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetSerializerFormat",
    info="read",
)
request_handlers.register_lazy(
    b"VersionedFileRepository.get_inventories",
    "breezy.bzr.smart.repository",
    "SmartServerRepositoryGetInventories",
    info="read",
)
request_handlers.register_lazy(
    b"rmdir", "breezy.bzr.smart.vfs", "RmdirRequest", info="semivfs"
)
request_handlers.register_lazy(
    b"stat", "breezy.bzr.smart.vfs", "StatRequest", info="read"
)
request_handlers.register_lazy(
    b"Transport.is_readonly",
    "breezy.bzr.smart.request",
    "SmartServerIsReadonly",
    info="read",
)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz rmbranch."""

from breezy import controldir
from breezy.tests import TestCaseWithTransport


class TestRemoveBranch(TestCaseWithTransport):
    def example_tree(self, path=".", format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree_contents([(path + "/hello", b"foo")])
        tree.add("hello")
        tree.commit(message="setup")
        self.build_tree_contents([(path + "/goodbye", b"baz")])
        tree.add("goodbye")
        tree.commit(message="setup")
        return tree

    def test_remove_local(self):
        # Remove a local branch.
        self.example_tree("a")
        self.run_bzr_error(
            ["Branch is active. Use --force to remove it.\n"], "rmbranch a"
        )
        self.run_bzr("rmbranch --force a")
        dir = controldir.ControlDir.open("a")
        self.assertFalse(dir.has_branch())
        self.assertPathExists("a/hello")
        self.assertPathExists("a/goodbye")

    def test_no_branch(self):
        # No branch in the current directory.
        self.make_repository("a")
        self.run_bzr_error(["Not a branch"], "rmbranch a")

    def test_no_tree(self):
        # removing the active branch is possible if there is no tree
        tree = self.example_tree("a")
        tree.controldir.destroy_workingtree()
        self.run_bzr("rmbranch", working_dir="a")
        dir = controldir.ControlDir.open("a")
        self.assertFalse(dir.has_branch())

    def test_no_arg(self):
        # location argument defaults to current directory
        self.example_tree("a")
        self.run_bzr_error(
            ["Branch is active. Use --force to remove it.\n"], "rmbranch a"
        )
        self.run_bzr("rmbranch --force", working_dir="a")
        dir = controldir.ControlDir.open("a")
        self.assertFalse(dir.has_branch())

    def test_remove_colo(self):
        # Remove a colocated branch.
        tree = self.example_tree("a")
        tree.controldir.create_branch(name="otherbranch")
        self.assertTrue(tree.controldir.has_branch("otherbranch"))
        self.run_bzr("rmbranch {},branch=otherbranch".format(tree.controldir.user_url))
        dir = controldir.ControlDir.open("a")
        self.assertFalse(dir.has_branch("otherbranch"))
        self.assertTrue(dir.has_branch())

    def test_remove_colo_directory(self):
        # Remove a colocated branch.
        tree = self.example_tree("a")
        tree.controldir.create_branch(name="otherbranch")
        self.assertTrue(tree.controldir.has_branch("otherbranch"))
        self.run_bzr("rmbranch otherbranch -d {}".format(tree.controldir.user_url))
        dir = controldir.ControlDir.open("a")
        self.assertFalse(dir.has_branch("otherbranch"))
        self.assertTrue(dir.has_branch())

    def test_remove_active_colo_branch(self):
        # Remove a colocated branch.
        dir = self.make_repository("a").controldir
        branch = dir.create_branch("otherbranch")
        branch.create_checkout("a")
        self.run_bzr_error(
            ["Branch is active. Use --force to remove it.\n"],
            "rmbranch otherbranch -d {}".format(branch.controldir.user_url),
        )
        self.assertTrue(dir.has_branch("otherbranch"))
        self.run_bzr(
            "rmbranch --force otherbranch -d {}".format(branch.controldir.user_url)
        )
        self.assertFalse(dir.has_branch("otherbranch"))



# Copyright (C) 2009, 2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for controldir implementations - push."""

from breezy.tests.per_controldir import TestCaseWithControlDir

from ...controldir import NoColocatedBranchSupport
from ...errors import LossyPushToSameVCS, NoSuchRevision, TagsNotSupported
from .. import TestNotApplicable


class TestPush(TestCaseWithControlDir):
    def create_simple_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a"])
        tree.add(["a"])
        rev_1 = tree.commit("one")
        return tree, rev_1

    def test_push_new_branch(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository("dir").controldir
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(dir.open_branch().base, result.target_branch.base)
        self.assertEqual(dir.open_branch().base, tree.branch.get_push_location())

    def test_push_to_colocated_active(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository("dir").controldir
        try:
            result = dir.push_branch(tree.branch, name="colo")
        except NoColocatedBranchSupport:
            raise TestNotApplicable("no colocated branch support")
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(dir.open_branch(name="colo").base, result.target_branch.base)
        self.assertEqual(
            dir.open_branch(name="colo").base, tree.branch.get_push_location()
        )

    def test_push_to_colocated_new_inactive(self):
        tree, rev_1 = self.create_simple_tree()
        target_tree = self.make_branch_and_tree("dir")
        rev_o = target_tree.commit("another")
        try:
            result = target_tree.branch.controldir.push_branch(tree.branch, name="colo")
        except NoColocatedBranchSupport:
            raise TestNotApplicable("no colocated branch support")
        target_branch = target_tree.branch.controldir.open_branch(name="colo")
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(target_tree.last_revision(), rev_o)
        self.assertEqual(target_tree.branch.last_revision(), rev_o)
        self.assertEqual(target_branch.base, result.target_branch.base)
        self.assertEqual(target_branch.base, tree.branch.get_push_location())
        self.assertNotEqual(target_branch.controldir.open_branch(name="").name, "colo")

    def test_push_to_colocated_existing_inactive(self):
        tree, rev_1 = self.create_simple_tree()
        target_tree = self.make_branch_and_tree("dir")
        rev_o = target_tree.commit("another")
        try:
            target_tree.branch.controldir.create_branch(name="colo")
        except NoColocatedBranchSupport:
            raise TestNotApplicable("no colocated branch support")

        try:
            result = target_tree.branch.controldir.push_branch(tree.branch, name="colo")
        except NoColocatedBranchSupport:
            raise TestNotApplicable("no colocated branch support")
        target_branch = target_tree.branch.controldir.open_branch(name="colo")
        self.assertEqual(tree.branch, result.source_branch)
        self.assertEqual(target_tree.last_revision(), rev_o)
        self.assertEqual(target_tree.branch.last_revision(), rev_o)
        self.assertEqual(target_branch.base, result.target_branch.base)
        self.assertEqual(target_branch.base, tree.branch.get_push_location())

    def test_push_no_such_revision(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository("dir").controldir
        self.assertRaises(
            NoSuchRevision, dir.push_branch, tree.branch, revision_id=b"idonotexist"
        )

    def test_push_new_branch_fetch_tags(self):
        builder = self.make_branch_builder("from")
        builder.start_series()
        rev_1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("filename", None, "file", b"content")),
            ],
        )
        rev_2 = builder.build_snapshot(
            [rev_1], [("modify", ("filename", b"new-content\n"))]
        )
        rev_3 = builder.build_snapshot(
            [rev_1], [("modify", ("filename", b"new-new-content\n"))]
        )
        builder.finish_series()
        branch = builder.get_branch()
        try:
            branch.tags.set_tag("atag", rev_2)
        except TagsNotSupported:
            raise TestNotApplicable("source format does not support tags")

        dir = self.make_repository("target").controldir
        branch.get_config().set_user_option("branch.fetch_tags", True)
        result = dir.push_branch(branch)
        self.assertEqual(
            {rev_1, rev_2, rev_3},
            set(result.source_branch.repository.all_revision_ids()),
        )
        self.assertEqual({"atag": rev_2}, result.source_branch.tags.get_tag_dict())

    def test_push_new_branch_lossy(self):
        tree, rev_1 = self.create_simple_tree()
        dir = self.make_repository("dir").controldir
        self.assertRaises(LossyPushToSameVCS, dir.push_branch, tree.branch, lossy=True)

    def test_push_new_empty(self):
        tree = self.make_branch_and_tree("tree")
        dir = self.make_repository("dir").controldir
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.branch.base, result.source_branch.base)
        self.assertEqual(dir.open_branch().base, result.target_branch.base)

    def test_push_incremental(self):
        tree, rev1 = self.create_simple_tree()
        dir = self.make_repository("dir").controldir
        dir.push_branch(tree.branch)
        self.build_tree(["tree/b"])
        tree.add(["b"])
        tree.commit("two")
        result = dir.push_branch(tree.branch)
        self.assertEqual(tree.last_revision(), result.branch_push_result.new_revid)
        self.assertEqual(2, result.branch_push_result.new_revno)
        self.assertEqual(tree.branch.base, result.source_branch.base)
        self.assertEqual(dir.open_branch().base, result.target_branch.base)

    def test_push_tag_selector(self):
        tree, rev1 = self.create_simple_tree()
        try:
            tree.branch.tags.set_tag("tag1", rev1)
        except TagsNotSupported:
            raise TestNotApplicable("tags not supported")
        tree.branch.tags.set_tag("tag2", rev1)
        dir = self.make_repository("dir").controldir
        dir.push_branch(tree.branch, tag_selector=lambda x: x == "tag1")
        self.assertEqual({"tag1": rev1}, dir.open_branch().tags.get_tag_dict())



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.switch."""

import os

from breezy import branch, errors, switch, tests, workingtree
from breezy import merge as _mod_merge


class TestSwitch(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.lightweight = True

    @staticmethod
    def _master_if_present(branch):
        master = branch.get_master_branch()
        if master:
            return master
        else:
            return branch

    def _setup_tree(self):
        tree = self.make_branch_and_tree("branch-1")
        self.build_tree(["branch-1/file-1"])
        tree.add("file-1")
        tree.commit("rev1")
        return tree

    def _setup_uncommitted(self, same_revision=False):
        tree = self._setup_tree()
        to_branch = tree.controldir.sprout("branch-2").open_branch()
        self.build_tree(["branch-1/file-2"])
        if not same_revision:
            tree.add("file-2")
            tree.remove("file-1")
            tree.commit("rev2")
        checkout = tree.branch.create_checkout("checkout", lightweight=self.lightweight)
        self.build_tree(["checkout/file-3"])
        checkout.add("file-3")
        return checkout, to_branch

    def test_switch_store_uncommitted(self):
        """Test switch updates tree and stores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")
        self.assertPathDoesNotExist("checkout/file-3")

    def test_switch_restore_uncommitted(self):
        """Test switch updates tree and restores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        old_branch = self._master_if_present(checkout.branch)
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        self.assertPathExists("checkout/file-3")
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        checkout = workingtree.WorkingTree.open("checkout")
        switch.switch(checkout.controldir, old_branch, store_uncommitted=True)
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        self.assertPathExists("checkout/file-3")

    def test_switch_restore_uncommitted_same_revision(self):
        """Test switch updates tree and restores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted(same_revision=True)
        old_branch = self._master_if_present(checkout.branch)
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        checkout = workingtree.WorkingTree.open("checkout")
        switch.switch(checkout.controldir, old_branch, store_uncommitted=True)
        self.assertPathExists("checkout/file-3")

    def test_switch_updates(self):
        """Test switch updates tree and keeps uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        switch.switch(checkout.controldir, to_branch)
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")
        self.assertPathExists("checkout/file-3")

    def test_switch_after_branch_moved(self):
        """Test switch after the branch is moved."""
        tree = self._setup_tree()
        checkout = tree.branch.create_checkout("checkout", lightweight=self.lightweight)
        self.build_tree(["branch-1/file-2"])
        tree.add("file-2")
        tree.remove("file-1")
        tree.commit("rev2")
        self.build_tree(["checkout/file-3"])
        checkout.add("file-3")
        # rename the branch on disk, the checkout object is now invalid.
        os.rename("branch-1", "branch-2")
        to_branch = branch.Branch.open("branch-2")
        # Check fails without --force
        err = self.assertRaises(
            (errors.CommandError, errors.NotBranchError),
            switch.switch,
            checkout.controldir,
            to_branch,
        )
        if isinstance(err, errors.CommandError):
            self.assertContainsRe(
                str(err),
                "Unable to connect to current master branch .*"
                "To switch anyway, use --force.",
            )
        switch.switch(checkout.controldir, to_branch, force=True)
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        self.assertPathExists("checkout/file-3")

    def test_switch_when_pending_merges(self):
        """Test graceful failure if pending merges are outstanding."""
        # Create 2 branches and a checkout
        tree = self._setup_tree()
        tree2 = tree.controldir.sprout("branch-2").open_workingtree()
        checkout = tree.branch.create_checkout("checkout", lightweight=self.lightweight)
        # Change tree2 and merge it into the checkout without committing
        self.build_tree(["branch-2/file-2"])
        tree2.add("file-2")
        tree2.commit("rev2")
        checkout.merge_from_branch(tree2.branch)
        # Check the error reporting is as expected
        err = self.assertRaises(
            errors.CommandError, switch.switch, checkout.controldir, tree2.branch
        )
        self.assertContainsRe(
            str(err), "Pending merges must be committed or reverted before using switch"
        )

    def test_switch_with_revision(self):
        """Test switch when a revision is given."""
        # Create a tree with 2 revisions
        tree = self.make_branch_and_tree("branch-1")
        self.build_tree(["branch-1/file-1"])
        tree.add("file-1")
        tree.commit(rev_id=b"rev1", message="rev1")
        self.build_tree(["branch-1/file-2"])
        tree.add("file-2")
        tree.commit(rev_id=b"rev2", message="rev2")
        # Check it out and switch to revision 1
        checkout = tree.branch.create_checkout("checkout", lightweight=self.lightweight)
        switch.switch(checkout.controldir, tree.branch, revision_id=b"rev1")
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")

    def test_switch_changing_root_id(self):
        tree = self._setup_tree()
        tree2 = self.make_branch_and_tree("tree-2")
        tree2.set_root_id(b"custom-root-id")
        self.build_tree(["tree-2/file-2"])
        tree2.add(["file-2"])
        tree2.commit("rev1b")
        checkout = tree.branch.create_checkout("checkout", lightweight=self.lightweight)
        switch.switch(checkout.controldir, tree2.branch)
        self.assertEqual(b"custom-root-id", tree2.path2id(""))

    def test_switch_configurable_file_merger(self):
        class DummyMerger(_mod_merge.ConfigurableFileMerger):
            name_prefix = "file"

        _mod_merge.Merger.hooks.install_named_hook(
            "merge_file_content", DummyMerger, "test factory"
        )
        foo = self.make_branch("foo")
        checkout = foo.create_checkout("checkout", lightweight=True)
        self.build_tree_contents([("checkout/file", b"a")])
        checkout.add("file")
        checkout.commit("a")
        bar = foo.controldir.sprout("bar").open_workingtree()
        self.build_tree_contents([("bar/file", b"b")])
        bar.commit("b")
        self.build_tree_contents([("checkout/file", b"c")])
        switch.switch(checkout.controldir, bar.branch)


class TestSwitchHeavyweight(TestSwitch):
    def setUp(self):
        super().setUp()
        self.lightweight = False

    def test_switch_with_local_commits(self):
        """Test switch complains about local commits unless --force given."""
        tree = self._setup_tree()
        to_branch = tree.controldir.sprout("branch-2").open_branch()
        self.build_tree(["branch-1/file-2"])
        tree.add("file-2")
        tree.remove("file-1")
        tree.commit("rev2")
        checkout = tree.branch.create_checkout("checkout")
        self.build_tree(["checkout/file-3"])
        checkout.add("file-3")
        checkout.commit(message="local only commit", local=True)
        self.build_tree(["checkout/file-4"])
        # Check the error reporting is as expected
        err = self.assertRaises(
            errors.CommandError, switch.switch, checkout.controldir, to_branch
        )
        self.assertContainsRe(
            str(err), "Cannot switch as local commits found in the checkout."
        )
        # Check all is ok when force is given
        self.assertPathDoesNotExist("checkout/file-1")
        self.assertPathExists("checkout/file-2")
        switch.switch(checkout.controldir, to_branch, force=True)
        self.assertPathExists("checkout/file-1")
        self.assertPathDoesNotExist("checkout/file-2")
        self.assertPathDoesNotExist("checkout/file-3")
        self.assertPathExists("checkout/file-4")
        # Check that the checkout is a true mirror of the bound branch
        self.assertEqual(
            to_branch.last_revision_info(), checkout.branch.last_revision_info()
        )



# Copyright (C) 2006, 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import branch, errors
from . import TestCaseWithTransport


class TestExtract(TestCaseWithTransport):
    def test_extract(self):
        self.build_tree(["a/", "a/b/", "a/b/c", "a/d"])
        wt = self.make_branch_and_tree("a", format="rich-root-pack")
        wt.add(["b", "b/c", "d"], ids=[b"b-id", b"c-id", b"d-id"])
        wt.commit("added files")
        b_wt = wt.extract("b")
        self.assertTrue(b_wt.is_versioned(""))
        if b_wt.supports_setting_file_ids():
            self.assertEqual(b"b-id", b_wt.path2id(""))
            self.assertEqual(b"c-id", b_wt.path2id("c"))
            self.assertEqual("c", b_wt.id2path(b"c-id"))
            self.assertRaises(errors.BzrError, wt.id2path, b"b-id")
        self.assertEqual(b_wt.basedir, wt.abspath("b"))
        self.assertEqual(wt.get_parent_ids(), b_wt.get_parent_ids())
        self.assertEqual(wt.branch.last_revision(), b_wt.branch.last_revision())

    def extract_in_checkout(self, a_branch):
        self.build_tree(["a/", "a/b/", "a/b/c/", "a/b/c/d"])
        wt = a_branch.create_checkout("a", lightweight=True)
        wt.add(["b", "b/c", "b/c/d"], ids=[b"b-id", b"c-id", b"d-id"])
        wt.commit("added files")
        return wt.extract("b")

    def test_extract_in_checkout(self):
        a_branch = self.make_branch("branch", format="rich-root-pack")
        self.extract_in_checkout(a_branch)
        b_branch = branch.Branch.open("branch/b")
        b_branch_ref = branch.Branch.open("a/b")
        self.assertEqual(b_branch.base, b_branch_ref.base)

    def test_extract_in_deep_checkout(self):
        a_branch = self.make_branch("branch", format="rich-root-pack")
        self.build_tree(["a/", "a/b/", "a/b/c/", "a/b/c/d/", "a/b/c/d/e"])
        wt = a_branch.create_checkout("a", lightweight=True)
        wt.add(
            ["b", "b/c", "b/c/d", "b/c/d/e/"], ids=[b"b-id", b"c-id", b"d-id", b"e-id"]
        )
        wt.commit("added files")
        wt.extract("b/c/d")
        b_branch = branch.Branch.open("branch/b/c/d")
        b_branch_ref = branch.Branch.open("a/b/c/d")
        self.assertEqual(b_branch.base, b_branch_ref.base)

    def test_bad_repo_format(self):
        repo = self.make_repository("branch", shared=True, format="knit")
        a_branch = repo.controldir.create_branch()
        self.assertRaises(errors.RootNotRich, self.extract_in_checkout, a_branch)

    def test_good_repo_format(self):
        repo = self.make_repository(
            "branch", shared=True, format="dirstate-with-subtree"
        )
        a_branch = repo.controldir.create_branch()
        wt_b = self.extract_in_checkout(a_branch)
        self.assertEqual(
            wt_b.branch.repository.controldir.transport.base,
            repo.controldir.transport.base,
        )
